package main

import (
	"log"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/errt"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo"

	mgo "github.com/globalsign/mgo"
	"github.com/streadway/amqp"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/alarm"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/config"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/encoding"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/encoding/json"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
)

type EngineAxe struct {
	canopsis.DefaultEngine
	Sessions          *canopsis.CpsSessions
	AlarmService      alarm.Service
	PeriodicalMongo   *mgo.Session
	PrintEventOnError bool
	jsonDecoder       encoding.Decoder
}

func (eng *EngineAxe) ConsumerChan() (<-chan amqp.Delivery, error) {
	msgs, err := eng.Sessions.AmqpSub.Consume(
		canopsis.AxeQueueName,    // queue
		canopsis.AxeConsumerName, // consumer
		false, // auto-ack
		false, // exclusive
		false, // no-local
		false, // no-wait
		nil,   // args
	)
	return msgs, err
}

// Initialize initialize the engine [sic]
func (eng *EngineAxe) Initialize() error {
	if err := eng.DefaultEngine.Initialize(); err != nil {
		return err
	}
	eng.PeriodicalMongo = eng.Sessions.Mongo.Copy()
	eng.jsonDecoder = json.NewDecoder()
	return nil
}

// PeriodicalProcess do actions on every beat
func (eng *EngineAxe) PeriodicalProcess() {
	eng.PeriodicalMongo.Refresh()
	as := alarm.NewService(eng.PeriodicalMongo, nil, config.GetConf().Alarm.EnableLastEventDate)
	// TODO replace with eng.AlarmService ?
	cfg := config.RefreshConf()
	baggotTime, err := time.ParseDuration(cfg.Alarm.BaggotTime)

	if err != nil {
		log.Printf("error: %v", err)
		return
	}

	eng.DebugPrint("closing alarms")
	closed, err := as.ResolveAlarms(baggotTime)
	if err != nil {
		log.Printf("error closing alarms: %v", err)
		eng.AskStop(canopsis.ExitEngine)
	}
	log.Printf("closed alarms: %d alarm(s)", closed)

	eng.DebugPrint("resolve snooze")
	snoozeResolved, err := as.ResolveSnoozes()
	if err != nil {
		log.Printf("error resolve snoozes: %v", err)
		eng.AskStop(canopsis.ExitEngine)
	}
	log.Printf("resolved snoozes: %d alarm(s)", snoozeResolved)

	eng.DebugPrint("resolve cancel")
	cancelResolved, err := as.ResolveCancels()
	if err != nil {
		log.Printf("error resolve cancels: %v", err)
		eng.AskStop(canopsis.ExitEngine)
	}
	log.Printf("resolved cancels: %d alarm(s)", cancelResolved)
}

func (e *EngineAxe) ackMsg(msg amqp.Delivery) error {
	return e.Sessions.AmqpSub.Ack(msg.DeliveryTag, false)
}

func (e *EngineAxe) processWorkerError(err error, msg amqp.Delivery) error {
	if err == nil {
		return nil
	}

	if !e.PrintEventOnError {
		log.Printf("event processing error: %v", err)
	} else {
		log.Printf("event processing error: %v\n%s", err, string(msg.Body))
	}

	switch err.(type) {
	case encoding.DecodingError:
		e.ackMsg(msg)
	case encoding.EncodingError:
		e.ackMsg(msg)
	case errt.UnmanagedEventError:
		e.ackMsg(msg)
	case errt.NotFound:
		e.ackMsg(msg)
	case errt.UnknownError:
		e.ackMsg(msg)
	case mongo.MongoError:
		e.ackMsg(msg)
		e.AskStop(canopsis.ExitEngine)
	default:
		log.Printf("error type: %T", err)
		e.AskStop(canopsis.ExitEngine)
	}

	return err
}

// WorkerProcess do actions on every incoming message
func (eng *EngineAxe) WorkerProcess(msg amqp.Delivery) {
	eng.DebugPrint("input: %s", string(msg.Body))

	var event types.Event
	err := eng.jsonDecoder.Decode(msg.Body, &event)
	if eng.processWorkerError(err, msg) != nil {
		return
	}

	eng.DebugPrint("unmarshaled: %+v", event)

	err = eng.AlarmService.ProcessAlarmEvent(&event)
	eng.processWorkerError(err, msg)

	eng.ackMsg(msg)
}
