package main

import (
	"fmt"
	"log"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/alarm"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/config"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/context"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/encoding"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/encoding/json"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/entity"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/event"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"git.canopsis.net/canopsis/go-revolution/lib/errt"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo"
	"github.com/streadway/amqp"
)

type EngineChe struct {
	canopsis.DefaultEngine
	ConsumeFromQueue      string
	PublishToQueue        string
	PublishToExchange     string
	EnrichFields          context.EnrichFields
	EnableEventProcessing bool
	EnableContextCreation bool
	EnableContextEnrich   bool
	EventDBAdapter        *event.Adapter
	EntityDBAdapter       *entity.DBAdapter
	channelPub            *amqp.Channel
	Sessions              *canopsis.CpsSessions
	EnrichmentCenter      context.EnrichmentCenter
	purge                 bool
	printEventOnError     bool
	JsonDecoder           encoding.Decoder
	JsonEncoder           encoding.Encoder
}

// NewEngineChe instanciate a new engine
func NewEngineChe(processEvent, createContext, enrichContext bool, publishQueue, excludeEnrich, includeEnrich string, purge bool, sessions *canopsis.CpsSessions, debug bool, printEventOnError bool) (canopsis.Engine, error) {
	if sessions == nil {
		return nil, fmt.Errorf("sessions cannot be nil")
	}

	if sessions.AmqpSub == nil {
		return nil, fmt.Errorf("sessions.AmqpSub cannot be nil")
	}

	alarmService := alarm.NewService(sessions.Mongo.Copy(), nil, config.GetConf().Alarm.EnableLastEventDate)
	entityAdapter := entity.NewAdapter(sessions.Mongo.Copy(), nil)

	evtdba := event.NewEventDBAdapter(&alarmService, &entityAdapter)

	defengine := canopsis.NewDefaultEngine(time.Second*60, true, true, sessions.AmqpSub)
	enrichmentCenter := context.NewSafeEnrichmentCenter(context.NewEnrichmentCenter(2000, enrichContext, &entityAdapter))
	defengine.Debug = debug
	engine := EngineChe{
		DefaultEngine:         defengine,
		EventDBAdapter:        &evtdba,
		EntityDBAdapter:       &entityAdapter,
		EnrichFields:          context.NewEnrichFields(includeEnrich, excludeEnrich),
		PublishToQueue:        publishQueue,
		EnableEventProcessing: processEvent,
		EnableContextCreation: createContext,
		EnableContextEnrich:   enrichContext,
		ConsumeFromQueue:      canopsis.CheQueueName,
		Sessions:              sessions,
		channelPub:            sessions.AmqpPub,
		EnrichmentCenter:      enrichmentCenter,
		purge:                 purge,
		printEventOnError:     printEventOnError,
		JsonDecoder:           json.NewDecoder(),
		JsonEncoder:           json.NewEncoder(),
	}

	return &engine, nil
}

func (e *EngineChe) ConsumerChan() (<-chan amqp.Delivery, error) {
	conf := config.GetConf()
	e.Sessions.AmqpSub.Qos(conf.Global.PrefetchCount, conf.Global.PrefetchSize, true)
	if e.purge {
		log.Printf("Purging queue %s", e.ConsumeFromQueue)
		_, err := e.Sessions.AmqpSub.QueuePurge(e.ConsumeFromQueue, false)
		if err != nil {
			return nil, fmt.Errorf("error while purging queue: %v", err)
		}
	}
	msgs, err := e.Sessions.AmqpSub.Consume(
		e.ConsumeFromQueue,       // queue
		canopsis.CheConsumerName, // consumer
		false, // auto-ack
		false, // exclusive
		false, // no-local
		false, // no-wait
		nil,   // args
	)
	return msgs, err
}

// Stop force bulk writes before stopping
func (e *EngineChe) Stop() {
	e.DebugPrint("flushing caches before exit")
	e.EnrichmentCenter.Flush(time.Now().Add(time.Hour * 1))
	e.EntityDBAdapter.FlushBulk()
	e.DefaultEngine.Stop()
}

func (e *EngineChe) PeriodicalProcess() {
	// EnrichContextFromEvent fill bulk cache for each updated entities
	e.DebugPrint("periodical bulk flush")
	e.EnrichmentCenter.Flush(time.Now().Add(time.Hour * 1))
	e.EntityDBAdapter.FlushBulk()
}

func (e *EngineChe) ackMsg(msg amqp.Delivery) error {
	return e.Sessions.AmqpSub.Ack(msg.DeliveryTag, false)
}

// processWorkerError read errors, ack amqp messages and stop the engine if needed
func (e *EngineChe) processWorkerError(err error, msg amqp.Delivery) error {
	if err == nil {
		return nil
	}

	if !e.printEventOnError {
		log.Printf("event processing error: %v", err)
	} else {
		log.Printf("event processing error: %v\n%s", err, string(msg.Body))
	}

	switch err.(type) {
	case encoding.DecodingError:
		e.ackMsg(msg)
	case encoding.EncodingError:
		e.ackMsg(msg)
	case mongo.MongoError:
		e.ackMsg(msg)
		e.AskStop(canopsis.ExitEngine)
	case errt.UnknownError:
		e.ackMsg(msg)
	default:
		e.ackMsg(msg)
		e.AskStop(canopsis.ExitEngine)
	}

	return err
}

func (e *EngineChe) WorkerProcess(msg amqp.Delivery) {
	var event types.Event
	err := e.JsonDecoder.Decode(msg.Body, &event)
	if e.processWorkerError(err, msg) != nil {
		return
	}

	event.InjectExtraInfos(msg.Body)

	e.DebugPrint("valid input event: %v", string(msg.Body))

	err = event.IsValid()
	if e.processWorkerError(err, msg) != nil {
		return
	}

	if e.EnableEventProcessing {
		if event.EventType == types.EventTypePBehavior {
			e.DebugPrint("event type Pbehavior. Passing it to next queue")
			e.processWorkerError(e.publishToNext(msg.Body), msg)
			// Ensure msg is always acked
			e.ackMsg(msg)
			return
		} else {
			event, err = e.cleanEnrichAndSend(event)
			if e.processWorkerError(err, msg) != nil {
				return
			}
		}
	}

	if e.EnableContextCreation || e.EnableContextEnrich {
		event, err = e.EnrichContextFromEvent(event)
		if e.processWorkerError(err, msg) != nil {
			return
		}
	}

	e.ackMsg(msg)
}

// EnrichContextFromEvent ...
func (e *EngineChe) EnrichContextFromEvent(event types.Event) (types.Event, error) {
	if e.EnableContextCreation && event.IsContextable() {
		event.Entity = e.EnrichmentCenter.Handle(event, e.EnrichFields)
	}
	return event, nil
}

// cLeanEnrichAndSend process an event and push to the new queue
func (e *EngineChe) cleanEnrichAndSend(event types.Event) (types.Event, error) {
	event.Format()

	event, report := e.EventDBAdapter.Enrich(event)
	if report.Error != nil {
		return event, report.Error
	}

	bevent, err := e.JsonEncoder.Encode(event)
	if err != nil {
		return event, err
	}

	e.DebugPrint("output event: %s", string(bevent))
	return event, e.publishToNext(bevent)
}

// publishToNext publish a received amqp message to another queue
func (e *EngineChe) publishToNext(event []byte) error {
	if e.PublishToQueue == "void" {
		return nil
	}

	return errt.NewIOError(e.channelPub.Publish(
		"",
		e.PublishToQueue,
		false,
		false,
		amqp.Publishing{
			ContentType:  "application/json", // this type is mandatory to avoid bad conversions into Python.
			Body:         event,
			DeliveryMode: amqp.Persistent,
		},
	))
}
