package main_test

import (
	"math/rand"
	"strconv"
	"testing"
	"time"

	. "git.canopsis.net/canopsis/go-revolution/cmd/engine-che"
	. "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/alarm"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/context"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/encoding/json"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/entity"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/event"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"github.com/globalsign/mgo/bson"
	. "github.com/smartystreets/goconvey/convey"
	"github.com/streadway/amqp"
)

type EngineTestChe struct {
	EngineChe
}

func (e *EngineTestChe) ConsumerChan() (<-chan amqp.Delivery, error) {
	ch := make(chan amqp.Delivery)
	return ch, nil
}

type testSendAlarm struct {
	Alarms []string
	Status []int
}

func (tsa *testSendAlarm) sendAlarmTest(alarmid string, state int) error {
	tsa.Alarms = append(tsa.Alarms, alarmid)
	tsa.Status = append(tsa.Status, state)
	return nil
}

func testNewEngineChe(sessions *CpsSessions) (*EngineTestChe, *testSendAlarm, error) {
	tsa := testSendAlarm{
		Alarms: make([]string, 0),
	}

	alarmService := alarm.NewService(sessions.Mongo.Copy(), nil, false)
	entityAdapter := entity.NewAdapter(sessions.Mongo.Copy(), nil)
	evtdba := event.NewEventDBAdapter(&alarmService, &entityAdapter)

	engine := EngineTestChe{
		EngineChe: EngineChe{
			DefaultEngine:         NewDefaultEngine(time.Second*60, true, true, sessions.AmqpSub),
			EventDBAdapter:        &evtdba,
			EntityDBAdapter:       &entityAdapter,
			PublishToQueue:        "Engine_event_filter",
			EnableEventProcessing: false,
			EnableContextCreation: true,
			EnableContextEnrich:   true,
			ConsumeFromQueue:      CheQueueName,
			Sessions:              sessions,
			EnrichmentCenter:      context.NewEnrichmentCenter(1000, true, &entityAdapter),
			JsonEncoder:           json.NewEncoder(),
			JsonDecoder:           json.NewDecoder(),
		},
	}
	_, err := engine.ConsumerChan()
	return &engine, &tsa, err
}

// Cleaning-up entity collection to evaluate insert/update perfs
func cleanUp(sessions *CpsSessions) {
	entityAdapter := entity.NewAdapter(sessions.Mongo.Copy(), nil)
	entityAdapter.Collection.Remove(bson.M{})
	sessions.Redis.FlushDB()
}

func TestInitializeChe(t *testing.T) {
	Convey("Setup", t, func() {
		sessions, err := NewCpsSessions(CpsSessionsConf{
			Mongo:   true,
			AmqpSub: true,
			Redis:   true,
		})
		So(err, ShouldBeNil)
		defer sessions.Close()
		So(sessions.Mongo, ShouldNotBeNil)
		So(sessions.AmqpSub, ShouldNotBeNil)
		So(sessions.Redis, ShouldNotBeNil)
		sessions.Redis.FlushDB()

		engine, _, err := testNewEngineChe(sessions)
		So(err, ShouldBeNil)

		Convey("Initialize will set correct warmups", func() {
			err := engine.Initialize()
			So(err, ShouldBeNil)

			r := sessions.Redis.Exists("connector:c1")
			So(r.Err(), ShouldBeNil)

			rttl := sessions.Redis.TTL("connector:c1")
			So(rttl.Err(), ShouldBeNil)
		})
	})
}

func TestWorkChe(t *testing.T) {
	Convey("Setup", t, func() {
		sessions, err := NewCpsSessions(CpsSessionsConf{
			Mongo:   true,
			AmqpSub: true,
			AmqpPub: true,
			Redis:   true,
		})
		So(err, ShouldBeNil)
		defer sessions.Close()
		So(sessions.Mongo, ShouldNotBeNil)
		So(sessions.Redis, ShouldNotBeNil)
		So(sessions.Amqp, ShouldNotBeNil)
		So(sessions.AmqpSub, ShouldNotBeNil)
		So(sessions.AmqpPub, ShouldNotBeNil)
		cleanUp(sessions)

		engine, _, err := testNewEngineChe(sessions)
		So(err, ShouldBeNil)
		err = engine.Initialize()
		So(err, ShouldBeNil)
		engine.EngineChe.EntityDBAdapter.Insert(types.Entity{ID: "force_collection_create"})

		Convey("Given a new event, it must create a context behind it", func() {
			So(err, ShouldBeNil)
			bevent := []byte(`{
					"event_type": "check",
					"source_type": "component",
					"state": 1,
					"connector": "air",
					"connector_name": "strike",
					"component": "electron",
					"spin": "1/2",
					"mass": "9.10938356(11)×10−31 kg"
				}`)
			event, err := types.NewEventFromJSON(bevent)
			So(err, ShouldBeNil)
			newEvent, err := engine.EnrichContextFromEvent(event)
			So(err, ShouldBeNil)
			So(newEvent.Connector, ShouldEqual, "air")
			So(newEvent.Component, ShouldEqual, "electron")
			engine.WorkerProcess(amqp.Delivery{
				Body: bevent,
			})
			engine.EnrichmentCenter.Flush(time.Now().Add(time.Hour * 1))
			err = engine.EntityDBAdapter.FlushBulk()
			So(err, ShouldBeNil)

			var entity types.Entity
			engine.EntityDBAdapter.Collection.Find(bson.M{"_id": "air/strike"}).One(&entity)
			So(entity, ShouldNotBeNil)
			So(entity.ID, ShouldNotEqual, "")
			So(entity.Type, ShouldEqual, types.EntityTypeConnector)

			entity = types.Entity{}
			engine.EntityDBAdapter.Collection.Find(bson.M{"_id": "electron"}).One(&entity)
			So(entity, ShouldNotBeNil)
			So(entity.Type, ShouldEqual, types.EntityTypeComponent)

			So(entity.Infos, ShouldContainKey, "spin")
			So(entity.Infos["spin"].Value, ShouldEqual, "1/2")
			So(entity.Infos, ShouldNotContainKey, "parity")

			Convey("Given a richer event on the same component", func() {
				bevent := []byte(`{
					"event_type": "check",
					"source_type": "component",
					"state": 1,
					"connector": "air",
					"connector_name": "strike",
					"component": "electron",
					"spin": "−1/2",
					"charge": "−1,602 × 10^−19 C"
				}`)
				event, err := types.NewEventFromJSON(bevent)
				So(err, ShouldBeNil)
				newEvent, err := engine.EnrichContextFromEvent(event)
				So(err, ShouldBeNil)
				So(newEvent.Component, ShouldEqual, "electron")
				engine.WorkerProcess(amqp.Delivery{
					Body: bevent,
				})
				engine.EnrichmentCenter.Flush(time.Now().Add(time.Hour * 1))
				err = engine.EntityDBAdapter.FlushBulk()
				So(err, ShouldBeNil)

				Convey("Then it must enrich the context without destroying old informations", func() {
					var entity2 types.Entity
					engine.EntityDBAdapter.Collection.Find(bson.M{"_id": "electron"}).One(&entity2)
					So(entity2, ShouldNotBeNil)
					So(entity2.Type, ShouldEqual, types.EntityTypeComponent)

					So(entity2.Infos, ShouldContainKey, "charge")
					charge := entity2.Infos["charge"]
					So(charge, ShouldNotBeNil)
					So(charge.Value, ShouldEqual, "−1,602 × 10^−19 C")

					So(entity2.Infos, ShouldContainKey, "spin")
					spin := entity2.Infos["spin"]
					So(spin, ShouldNotBeNil)
					So(spin.Value, ShouldEqual, "−1/2")

					So(entity2.Infos, ShouldContainKey, "mass")
					mass := entity2.Infos["mass"]
					So(mass, ShouldNotBeNil)
					So(mass.Value, ShouldEqual, "9.10938356(11)×10−31 kg")
				})
			})
		})

		Convey("Given a bad event, it must not create a context", func() {
			sessions.Redis.FlushDB()
			bevent := []byte(`{"stevie": "wonder"}`)
			event, err := types.NewEventFromJSON(bevent)
			So(err, ShouldBeNil)
			event.EventType = ""

			newEvent, err := engine.EnrichContextFromEvent(event)
			So(err, ShouldBeNil)
			So(newEvent.Entity, ShouldBeNil)
		})
	})
}

func TestEmptyEntityInfos(t *testing.T) {
	Convey("Given an entity without infos", t, func() {
		info := types.NewInfo("talking", "", "heads")
		entity := types.NewEntity("the", "great", "curve", nil, types.NewDedupList(), types.NewDedupList())
		So(entity.Infos, ShouldBeNil)

		Convey("Then we should initialise infos to populate it", func() {
			_, exists := entity.Infos[info.Name]
			So(exists, ShouldBeFalse)

			entity.Infos = make(map[string]types.Info, 0)
			_, exists = entity.Infos[info.Name]
			So(exists, ShouldBeFalse)
			entity.Infos[info.Name] = info
			_, exists = entity.Infos[info.Name]
			So(exists, ShouldBeTrue)
		})
	})
}

func BenchmarkWorkChe(b *testing.B) {
	sessions, err := NewCpsSessions(CpsSessionsConf{
		Mongo:   true,
		AmqpPub: true,
		AmqpSub: true,
		Redis:   true,
	})
	if err != nil {
		b.Error(err)
	}
	defer sessions.Close()
	sessions.Redis.FlushDB()
	engine, _, err := testNewEngineChe(sessions)
	if err != nil {
		b.Error(err)
	}

	engine.Initialize()
	cleanUp(sessions)

	conn := 1
	comp := 1
	for i := 1; i < b.N; i++ {
		if i%10 == 0 {
			if i%100 == 0 {
				conn = conn + 1
			} else {
				comp = comp + 1
			}
		}

		sevent := `{
			"event_type": "check",
			"source_type": "resource",
			"connector": "perceval_` + strconv.Itoa(conn) + `",
			"connector_name": "pellinor",
			"component": "gallois_` + strconv.Itoa(comp) + `",
			"resource": "benchmark_work_che_` + strconv.Itoa(i) + `",
			"state": ` + strconv.Itoa(rand.Intn(3)) + `,
			"output": "Selon comme on est tourné, ça change tout !"
		}`

		engine.WorkerProcess(amqp.Delivery{
			Body: []byte(sevent),
		})
	}
}
