package main

import (
	"errors"
	"fmt"
	"log"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/config"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/heartbeat"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"github.com/streadway/amqp"
)

type EngineHeartBeat struct {
	cps.DefaultEngine
	Pub               *amqp.Channel
	Sessions          *cps.CpsSessions
	Cfg               *config.Conf
	HeartBeatItems    []heartbeat.Item
	regheartBeatItems map[string]int
	SendAlarmFunc     func(string, int) error
}

// AddHeartBeatItem avoid to register duplicate HeartBeatItem-s in the Engine
func (eng *EngineHeartBeat) AddHeartBeatItem(li heartbeat.Item) error {
	_, exists := eng.regheartBeatItems[li.ID()]
	if !exists {
		eng.HeartBeatItems = append(eng.HeartBeatItems, li)
	} else {
		return errors.New("this HeartBeatItem already exists in the set")
	}
	return nil
}

func (eng *EngineHeartBeat) ConsumerChan() (<-chan amqp.Delivery, error) {
	_, err := eng.Sub.QueueDeclare(
		cps.HeartBeatQueueName,
		true,  // durable
		false, // autodelete
		false, // exclusive
		false, // nowait
		nil,   // args
	)
	if err != nil {
		return nil, err
	}
	err = eng.Sub.QueueBind(
		cps.HeartBeatQueueName,
		"#",
		"canopsis.events",
		false,
		nil,
	)
	if err != nil {
		return nil, err
	}

	return eng.Sub.Consume(
		cps.HeartBeatQueueName,    // queue
		cps.HeartBeatConsumerName, // consumer
		false, // auto-ack
		false, // exclusive
		false, // no-local
		false, // no-wait
		nil,   // args
	)
}

func (eng *EngineHeartBeat) WorkerProcess(msg amqp.Delivery) {
	var ievent types.GenericEvent
	err := ievent.JSONUnmarshal(msg.Body)
	if err != nil {
		log.Printf("invalid json: %v\n", err)
		eng.Sub.Ack(msg.DeliveryTag, false)
		return
	}
	_, err = eng.workHeartBeat(ievent)
	if err != nil {
		log.Printf("error on event: %v\n", err)
	}
	eng.Sub.Ack(msg.DeliveryTag, false)
}

func (eng *EngineHeartBeat) sendalarm(alarmid string, state int) error {
	msg := `{"connector":"heartbeat","connector_name":"heartbeat","event_type":"check","component":"heartbeats","source_type":"resource", "resource":"`
	msg = fmt.Sprintf(`%s%s", "state":%d}`, msg, alarmid, state)
	return eng.Pub.Publish(
		cps.HeartBeatExchangeName,
		"#",
		false,
		false,
		amqp.Publishing{
			Body:        []byte(msg),
			ContentType: "application/json",
		},
	)
}

func (eng *EngineHeartBeat) PeriodicalProcess() {
	log.Print("periodical process")
	for _, heartBeatItem := range eng.HeartBeatItems {
		id := heartBeatItem.ID()
		if id == "" {
			continue
		}
		result := eng.Sessions.Redis.Exists(id)
		if result.Val() == 0 {
			err := eng.SendAlarmFunc(id, types.AlarmStateCritical)
			if err == nil {
				eng.Sessions.Redis.Set("alarm:"+id, 0, 0)
			}
		} else {
			_ = eng.SendAlarmFunc(id, types.AlarmStateOK)
		}
	}

}

func (eng *EngineHeartBeat) warmup(li heartbeat.Item) error {
	res := eng.Sessions.Redis.Exists(li.ID())
	if res.Err() == nil && res.Val() == 0 {
		setr := eng.Sessions.Redis.Set(li.ID(), 0, li.MaxDuration)
		return setr.Err()
	}

	return nil
}

func (eng *EngineHeartBeat) Initialize() error {
	if err := eng.DefaultEngine.Initialize(); err != nil {
		return err
	}

	eng.HeartBeatItems = make([]heartbeat.Item, 0)
	for _, cfgli := range eng.Cfg.HeartBeat.Items {
		li, err := cfgli.ToHeartBeatItem()
		if err != nil {
			return err
		}
		err = eng.AddHeartBeatItem(li)
		if err != nil {
			return err
		}
		eng.warmup(li)
	}
	log.Printf("loaded %d rule(s)", len(eng.HeartBeatItems))
	return nil
}

func (eng *EngineHeartBeat) resolveAlarm(id string) error {
	return eng.SendAlarmFunc(id, types.AlarmStateOK)
}

func (eng *EngineHeartBeat) workHeartBeat(ievent types.GenericEvent) ([]int, error) {
	itemCounts := make([]int, 0)
	for itemCount, lm := range eng.HeartBeatItems {
		id, err := ievent.PartialID(lm)
		if err != nil || id != lm.ID() {
			continue
		}

		// Set expiration date to lm.MaxDuration so when the
		// beat processing works, it only checks if the key
		// exists or no.
		cmdstatus := eng.Sessions.Redis.Set(id, nil, lm.MaxDuration)
		if cmdstatus.Err() != nil {
			log.Printf("redis set \"%s\" error: %v", id, cmdstatus.Err())
			continue
		}

		// Check for an alarm that we must solve.
		alarmID := "alarm:" + id
		cmdint := eng.Sessions.Redis.Exists(alarmID)
		if cmdint.Err() != nil {
			log.Printf("redis exists \"%s\" error: %v", alarmID, cmdint.Err())
			continue
		}

		if cmdint.Val() != 0 {
			log.Printf("resolution: %s", alarmID)
			cmddel := eng.Sessions.Redis.Del(alarmID)
			if cmddel.Err() != nil {
				log.Printf("redis del \"%s\" error: %v", alarmID, cmddel.Err())
			}
			err := eng.resolveAlarm(id)
			if err != nil {
				log.Printf("error resolving alarm on %s: %v", id, err)
			}
		}

		itemCounts = append(itemCounts, itemCount)
	}

	return itemCounts, nil
}
