package main

import (
	"encoding/json"
	"fmt"
	"log"
	"time"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/config"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/stat"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"git.canopsis.net/canopsis/go-revolution/lib/influx/bulk"
	mgobulk "git.canopsis.net/canopsis/go-revolution/lib/mongo/bulk"
	influxmod "github.com/influxdata/influxdb/client/v2"
	"github.com/streadway/amqp"
)

// EngineStat regroup all needed parameters for the engine to run
type EngineStat struct {
	cps.DefaultEngine
	Sessions *cps.CpsSessions
	Cfg      *config.Conf
	statBulk bulk.Bulk
	mgoBulk  mgobulk.Bulk
}

// ConsumerChan consume an amqp queue
func (eng *EngineStat) ConsumerChan() (<-chan amqp.Delivery, error) {
	msgs, err := eng.Sub.Consume(
		cps.StatsQueueName,    // queue
		cps.StatsConsumerName, // consumer
		false, // auto-ack
		false, // exclusive
		false, // no-local
		false, // no-wait
		nil,   // args
	)
	if err != nil {
		return nil, fmt.Errorf("stat consumer chan: %v", err)
	}
	return msgs, nil
}

func (eng *EngineStat) resetBulk() error {
	var err error
	batchConfig := influxmod.BatchPointsConfig{
		Database:  cps.StatsDatabase,
		Precision: "s",
	}

	unsafeBulk, err := bulk.New(1000, eng.Sessions.Influx, batchConfig)

	if err != nil {
		return fmt.Errorf("Engine stat: resetBatchPoints: %v", err)
	}

	eng.statBulk = bulk.NewSafe(unsafeBulk)

	return nil
}

func (eng *EngineStat) resetMgoBulk() error {
	collection := eng.Sessions.Mongo.DB(cps.DbName).C("object")
	unsafe := mgobulk.New(collection, 1000)
	safe := mgobulk.NewSafe(unsafe)
	eng.mgoBulk = safe
	return nil
}

// Initialize do specific initialization for the engine
func (eng *EngineStat) Initialize() error {
	if err := eng.DefaultEngine.Initialize(); err != nil {
		return err
	}

	if err := eng.resetBulk(); err != nil {
		return err
	}

	if err := eng.resetMgoBulk(); err != nil {
		return err
	}

	return nil
}

// WorkerProcess handle arriving message from amqp
func (eng *EngineStat) WorkerProcess(msg amqp.Delivery) {
	var event types.Event
	err := json.Unmarshal(msg.Body, &event)
	if err != nil {
		log.Printf("invalid json: %v", err)
		eng.DebugPrint("erroneous event: %s", string(msg.Body))
		eng.Sub.Ack(msg.DeliveryTag, false)
		return
	}
	eng.DebugPrint("incoming event for %s", event.GetEID())

	statItem := eng.WorkStat(event)
	if statItem != nil {
		if err := eng.publishOnInflux(*statItem); err != nil {
			log.Printf("fail to publish %v", err)
		}
	}

	if event.Author != "" && event.EventType != "check" {
		eng.DebugPrint("insert log author: %+v", event)
		err := eng.insertLogAuthor(event)
		if err != nil {
			log.Printf("fail to insert log author: %v", err)
			eng.Sub.Ack(msg.DeliveryTag, false)
			return
		}
	}

	eng.Sub.Ack(msg.DeliveryTag, false)
}

// PeriodicalProcess process stats periodicaly (Needed function for all engines)
func (eng *EngineStat) PeriodicalProcess() {
	if err := eng.statBulk.Perform(); err != nil {
		log.Printf("periodical stat bulk error: %v", err)
		eng.resetBulk()
	}

	_, err := eng.mgoBulk.PerformInserts()
	if err != nil {
		log.Printf("periodical mgo bulk error: %v", err)
	}
}

// WorkStat process an Event and push stats to influx
func (eng *EngineStat) WorkStat(event types.Event) *stat.Item {
	if event.EventType != types.EventTypeCheck {
		return nil
	}

	isOk := event.State == 0
	item := stat.NewItem(
		event.Connector,
		event.ConnectorName,
		event.Component,
		event.Resource,
		event.GetEID(),
		isOk,
	)
	return &item
}

// LogAuthor is a structure that describes the mongodb log author document.
type LogAuthor struct {
	Author      string `bson:"author"`
	EventType   string `bson:"action"`
	Eid         string `bson:"cible"`
	Date        int64  `bson:"date"`
	CRecordType string `bson:"crecord_type"`
}

// insertLogAuthor log in mongodb actions made by users from events
func (eng *EngineStat) insertLogAuthor(event types.Event) error {
	log := LogAuthor{
		Author:      event.Author,
		EventType:   event.EventType,
		Eid:         event.GetEID(),
		Date:        time.Now().Unix(),
		CRecordType: "logauthor",
	}
	op := mgobulk.NewInsert(log)
	_, err := eng.mgoBulk.AddInsert(op)
	return err
}

// publishOnInflux push event statistic data to influx
func (eng *EngineStat) publishOnInflux(item stat.Item) error {
	tags := map[string]string{
		"connector":      item.Connector,
		"connector_name": item.ConnectorName,
		"component":      item.Component,
		"resource":       item.Resource,
		"eid":            item.ID,
	}
	fields := map[string]interface{}{
		"ok": item.EventOk,
		"ko": item.EventKo,
	}

	pointOp := bulk.PointOp{
		Name:   cps.StatsMeasurement,
		Tags:   tags,
		Fields: fields,
		Time:   time.Now(),
	}

	if err := eng.statBulk.AddPoints(pointOp); err != nil {
		return fmt.Errorf("stat.AddPoints: %v", err)
	}

	return nil
}
