package main

import (
	"encoding/json"
	"fmt"
	"log"
	"math"
	"strconv"
	"time"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"github.com/streadway/amqp"
)

type Feeder struct {
	flags Flags
	s     *cps.CpsSessions
}

func (f *Feeder) getDirtyEvent(state, Ci, ci, ri int64) types.Event {
	return types.Event{
		Connector:     "feeder2",
		ConnectorName: "feeder2_inst" + strconv.Itoa(int(Ci)),
		Component:     "feeder2_" + strconv.Itoa(int(ci)),
		Resource:      "feeder2_" + strconv.Itoa(int(ri)),
		State:         types.CpsNumber(state),
		SourceType:    types.SourceTypeResource,
		EventType:     types.EventTypeCheck,
	}
}

func (f *Feeder) getCompatEvent(state, Ci, ci, ri int64) types.Event {
	st := types.CpsNumber(1)
	return types.Event{
		Connector:     "feeder2",
		ConnectorName: "feeder2_inst" + strconv.Itoa(int(Ci)),
		Component:     "feeder2_" + strconv.Itoa(int(ci)),
		Resource:      "feeder2_" + strconv.Itoa(int(ri)),
		State:         types.CpsNumber(state),
		StateType:     &st,
		SourceType:    types.SourceTypeResource,
		EventType:     types.EventTypeCheck,
		Timestamp:     types.CpsTime{time.Now()},
	}
}

func (f *Feeder) sendBytes(content []byte, rk string) error {
	return f.s.AmqpPub.Publish(
		f.flags.ExchangeName,
		rk,
		false,
		false,
		amqp.Publishing{
			Body:        content,
			ContentType: "application/json",
		},
	)
}

func (f *Feeder) send(state, Ci, ci, ri int64) error {
	var evt types.Event
	if f.flags.DirtyEvent {
		evt = f.getDirtyEvent(state, Ci, ci, ri)
	} else {
		evt = f.getCompatEvent(state, Ci, ci, ri)
	}
	bevt, _ := json.Marshal(evt)
	return f.sendBytes(bevt, evt.GetCompatRK())
}

func (f *Feeder) adjust(target float64, sent, tsent int64) int64 {
	eps := float64(sent * time.Second.Nanoseconds() / tsent)
	variance := math.Abs(target-eps) * 100.0 / float64(target)
	adj := int64(0)

	if variance > 1 {
		if eps < target {
			adj = -tsent / 100
		} else {
			adj = tsent / 100
		}
	}

	return adj
}

func (f *Feeder) setupAmqp() error {
	s, err := cps.NewCpsSessions(cps.CpsSessionsConf{
		AmqpPub: true,
	})

	if err != nil {
		return fmt.Errorf("sessions: %v", err)
	}

	err = s.AmqpPub.Confirm(false)
	if err != nil {
		return fmt.Errorf("confirm: %v", err)
	}

	f.s = s

	return nil
}

func (f *Feeder) Run() error {
	var err error
	switch f.flags.Mode {
	case "file":
		if f.flags.PubHTTP {
			f.modeSendEventHTTP()
		} else if f.flags.PubAMQP {
			err = f.modeSendEvent()
		}
	case "feeder":
		err = f.modeFeeder()
	default:
		err = fmt.Errorf("Unknown mode \"%s\": please check help (-h)", f.flags.Mode)
	}

	return err
}

func NewFeeder() (*Feeder, error) {
	f := Feeder{}
	if err := f.flags.ParseArgs(); err != nil {
		return nil, err
	}

	if f.flags.Version {
		cps.PrintVersionExit()
	}

	return &f, nil
}

func main() {
	feeder, err := NewFeeder()
	if err != nil {
		log.Fatalf("feeder init error: %v", err)
	}

	if err = feeder.Run(); err != nil {
		log.Fatalf("feeder run error: %v", err)
	}
}
