package main

import (
	"flag"
	"fmt"
	"io/ioutil"
	"os"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/utils"

	toml "github.com/BurntSushi/toml"
)

const (
	ErrGeneral    = 1
	ErrRabbitInit = 2

	DefaultCfgFile = "initialisation.toml"
	FlagUsageConf  = "The configuration file used to initialize Canopsis."
)

func main() {
	var confFile string
	flag.StringVar(&confFile, "conf", DefaultCfgFile, FlagUsageConf)
	flag.Parse()

	data, err := ioutil.ReadFile(confFile)
	if err != nil {
		fmt.Printf("Error %+v\n", err)
		os.Exit(1)
	}

	var conf Conf
	if _, err := toml.Decode(string(data), &conf); err != nil {
		fmt.Printf("Error %+v\n", err)
		os.Exit(1)
	}

	scfg := canopsis.CpsSessionsConf{
		Amqp: true,
	}
	sessions, err := canopsis.GracefullStart(scfg)
	utils.FailOnError(err, "Failed to open one of required sessions")
	defer sessions.Close()

	ch, err := sessions.Amqp.Channel()
	utils.FailOnError(err, "Failed to create channel amqp")

	fmt.Printf("Initialising RabbitMQ exchanges : \n")
	for _, exchange := range conf.RabbitMQ.Exchanges {

		err := ch.ExchangeDeclare(exchange.Name,
			exchange.Kind,
			exchange.Durable,
			exchange.AutoDelete,
			exchange.Internal,
			exchange.NoWait,
			exchange.Args)
		if err != nil {
			fmt.Fprintf(os.Stderr, "Can not initialise exchange %s : %v\n",
				exchange.Name, err)
			os.Exit(ErrRabbitInit)
		}
		fmt.Printf("\tExchange \"%s\" created.\n", exchange.Name)
	}

	fmt.Printf("\nInitialising RabbitMQ queues : \n")
	for _, queue := range conf.RabbitMQ.Queues {

		_, err := ch.QueueDeclare(queue.Name,
			queue.Durable,
			queue.AutoDelete,
			queue.Exclusive,
			queue.NoWait,
			queue.Args)

		if err != nil {
			fmt.Fprintf(os.Stderr, "Can not initialise queue %s : %v\n",
				queue.Name, err)
			os.Exit(ErrRabbitInit)
		}

		fmt.Printf("\tQueue \"%s\" created.\n", queue.Name)

		if queue.Bind != nil {

			err := ch.QueueBind(queue.Name,
				queue.Bind.Key,
				queue.Bind.Exchange,
				queue.Bind.NoWait,
				queue.Bind.Args)

			if err != nil {
				fmt.Fprintf(os.Stderr, "Can not bind queue %s to exchange %s %v\n",
					queue.Name, queue.Bind.Exchange, err)
				os.Exit(ErrRabbitInit)
			}
			fmt.Printf("\t\t\"%s\" bind to \"%s\" exchange with \"%s\" routing key.\n",
				queue.Name,
				queue.Bind.Exchange,
				queue.Bind.Key)
		}

	}

}
