package main

import (
	"flag"
	"fmt"
	"log"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/amqp"
	"git.canopsis.net/canopsis/go-revolution/lib/influx"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo"
	"git.canopsis.net/canopsis/go-revolution/lib/redis"
)

func timeout(t time.Duration) {
	time.Sleep(t)
	log.Fatalf("timeout expired after %s", t.String())
}

func check(checker func() error, name string, retries int) error {
	infinite := false

	if retries == 0 {
		infinite = true
	}

	for retry := 0; retry < retries || infinite; retry++ {
		err := checker()
		if err != nil {
			log.Printf("%v: %v/%v: %v", name, retry+1, retries, err)
			time.Sleep(time.Second * 1)
		} else {
			return nil
		}
	}

	return fmt.Errorf("check %v failed after %v tries", name, retries)
}

func abort(err error) {
	if err != nil {
		log.Fatalf("hard failure: %v", err)
	}
}

func checkRedis() error {
	_, err := redis.NewSession(0)
	return err
}

func checkMongo() error {
	_, err := mongo.NewSession(mongo.Timeout)
	return err
}

func checkAMQP() error {
	_, err := amqp.NewSession()
	return err
}

func checkInflux() error {
	_, err := influx.NewSession()
	return err
}

func main() {
	var flagRetries int
	var flagTimeout time.Duration
	flag.IntVar(&flagRetries, "retries", 10, "number of retries per check. if 0, infinite number of retries")
	flag.DurationVar(&flagTimeout, "timeout", time.Second*60, "timeout after given duration. never timeout if 0s")

	flag.Parse()

	if flagTimeout != time.Second*0 {
		go timeout(flagTimeout)
	}

	abort(check(checkRedis, "redis", flagRetries))
	abort(check(checkMongo, "mongo", flagRetries))
	abort(check(checkAMQP, "amqp", flagRetries))
	abort(check(checkInflux, "influx", flagRetries))

	log.Println("ready!")
}
