package alarm

import (
	gogob "encoding/gob"
	"fmt"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/cache"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/encoding"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/encoding/gob"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	cpsredis "git.canopsis.net/canopsis/go-revolution/lib/redis"
	"git.canopsis.net/canopsis/go-revolution/lib/utils"
	"github.com/go-redis/redis"
)

// Cache uses Redis and a local map for alarms.
type Cache struct {
	backend *redis.Client
	decoder encoding.Decoder
	encoder encoding.Encoder
}

// NewBackend handle cache connection
func NewBackend() (*redis.Client, error) {
	return cpsredis.NewSession(cpsredis.CacheAlarm)
}

// NewCache creates the local cache and remote cache access for alarms.
// You may want to wrap with CacheThreadSafe.
func NewCache(redis *redis.Client) cache.Cache {
	if redis == nil {
		nredis, err := NewBackend()
		utils.FailOnError(err, "alarm cache redis")
		redis = nredis
	}
	gogob.Register(types.Alarm{})
	c := Cache{
		backend: redis,
		decoder: gob.NewDecoder(),
		encoder: gob.NewEncoder(),
	}
	return &c
}

// Flush resets local and remote caches
func (c *Cache) Flush() error {
	return c.backend.FlushDB().Err()
}

// Get an Alarm from cache if available, returns nil, false if not.
// Example:
//
// val, err := cache.Get(id)
// alarm := val.(Alarm)
func (c *Cache) Get(alarmid string) (interface{}, bool) {
	r := c.backend.Get(alarmid)

	if r.Err() == redis.Nil {
		return nil, false
	}

	b, err := r.Bytes()
	if err != nil {
		c.Drop(alarmid)
		return nil, false
	}
	var alarm types.Alarm
	err = c.decoder.Decode(b, &alarm)

	if err != nil {
		switch err.(type) {
		case encoding.DecodingError:
			c.Drop(alarmid)
			return nil, false
		default:
			return nil, false
		}
	}

	return alarm, true
}

// Drop alarms in local and remote cache. Use Alarm.AlarmID() for ids.
func (c *Cache) Drop(alarmids ...string) error {
	return c.backend.Del(alarmids...).Err()
}

// Set the local and remote cache with this alarm.
func (c *Cache) Set(alarm cache.Cacheable) error {
	b, err := c.encoder.Encode(alarm.(types.Alarm))
	if err != nil {
		switch err.(type) {
		case encoding.EncodingError:
			return encoding.NewDecodingError(fmt.Errorf("alarm cache: %v", err))
		default:
			return fmt.Errorf("alarm cache unexpected: %v", err)
		}
	}
	return c.backend.Set(alarm.CacheID(), b, time.Hour*4).Err()
}
