package alarm

import (
	"fmt"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo/bulk"
	mgo "github.com/globalsign/mgo"
	"github.com/globalsign/mgo/bson"
)

// AlarmCollectionName is where we store alarms
const (
	AlarmCollectionName = "periodical_alarm"
)

// Adapter allows you to manipulate alarms in database
type Adapter struct {
	Collection *mgo.Collection
	mbulk      bulk.Bulk
}

// NewAdapter gives the correct mongo alarm adapter.
func NewAdapter(session *mgo.Session) Adapter {
	collection := session.DB(cps.DbName).C(AlarmCollectionName)

	return Adapter{
		Collection: collection,
		mbulk:      bulk.NewSafe(bulk.New(collection, cps.DefaultBulkSize)),
	}
}

// Insert insert an alarm
func (a *Adapter) Insert(alarm types.Alarm) error {
	return mongo.WrapError(a.Collection.Insert(alarm))
}

// UpdateId update an alarm
func (a *Adapter) Update(alarm types.Alarm) error {
	return mongo.WrapError(a.Collection.UpdateId(alarm.ID, alarm))
}

// RemoveId remove alarm from a query
func (a *Adapter) Remove(query bson.M) error {
	return mongo.WrapError(a.Collection.Remove(query))
}

// RemoveId remove an alarm from its id
func (a *Adapter) RemoveId(id string) error {
	return mongo.WrapError(a.Collection.RemoveId(id))
}

// GetUnresolved returns all alarms that have v.resolved to null or absent
// field.This function does not use cache.
func (a *Adapter) GetUnresolved() ([]types.Alarm, error) {
	alarms := make([]types.Alarm, 0)
	// no need to check for the field to exist.
	err := mongo.WrapError(a.Collection.Find(bson.M{"v.resolved": nil}).All(&alarms))
	return alarms, err
}

// GetAlarmsWithSnoozeMark returns all alarms where v.snooze is not null
func (a *Adapter) GetAlarmsWithSnoozeMark() ([]types.Alarm, error) {
	alarms := make([]types.Alarm, 0)
	filter := bson.M{"v.snooze": bson.M{"$ne": nil}}
	err := mongo.WrapError(a.Collection.Find(filter).All(&alarms))
	return alarms, err
}

// GetAlarmsWithCancelMark returns all alarms where v.cancel is not null
func (a *Adapter) GetAlarmsWithCancelMark() ([]types.Alarm, error) {
	alarms := make([]types.Alarm, 0)
	filter := bson.M{"v.canceled": bson.M{"$ne": nil}}
	err := mongo.WrapError(a.Collection.Find(filter).All(&alarms))
	return alarms, err
}

// GetByEntityID finds all alarms with an entity id. This function does not
// use cache.
func (a *Adapter) GetByEntityID(id string) ([]types.Alarm, error) {
	al := []types.Alarm{}
	err := mongo.WrapError(a.Collection.Find(bson.M{"d": id}).All(&al))
	return al, err
}

// GetOpenedAlarm find one opened alarm with his entity id.
// Note : a control is added to prevent fetching future alarms.
func (a *Adapter) GetOpenedAlarm(connector, connectorName, id string) (types.Alarm, error) {
	al := types.Alarm{}
	query := bson.M{
		"d":                id,
		"v.connector":      connector,
		"v.connector_name": connectorName,
		"$or": []bson.M{
			bson.M{"v.resolved": nil},
			bson.M{"v.resolved": bson.M{"$exists": false}},
		},
	}
	err := mongo.WrapError(a.Collection.Find(query).One(&al))

	return al, err
}

// GetNotAckNotResolvedAlarms find all alarms which are not ack, not resolved and
// associated to an event. An alarm is associated to an event if his entities
// is link to the component of the given event.
func (a *Adapter) GetNotAckNotResolvedAlarms(component string) ([]types.Alarm, error) {
	req := bson.M{
		"v.component": component,
		"v.state":     bson.M{"$ne": "0"},
		"v.ack":       nil,
	}

	alarms := make([]types.Alarm, 0)
	err := mongo.WrapError(a.Collection.Find(req).All(&alarms))

	return alarms, err
}

// MassUpdate will update alarms with a mgo.Bulk object, by slices of max 1000
// alarms.
func (a *Adapter) MassUpdate(alarms []types.Alarm) error {
	for _, alarm := range alarms {
		op := bulk.NewUpdate(bson.M{"_id": alarm.ID}, alarm)
		_, err := a.mbulk.AddUpdate(op)
		if err != nil {
			return fmt.Errorf("alarm.Adapter.MassUpdate: %v", err)
		}
	}

	_, err := a.mbulk.PerformUpdates()
	if err != nil {
		return fmt.Errorf("alarm.Adapter.MussUpdate - flush: %v", err)
	}

	return nil
}
