package alarm

import (
	"fmt"
	"log"
	"time"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/cache"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"git.canopsis.net/canopsis/go-revolution/lib/errt"
	mgo "github.com/globalsign/mgo"
	"github.com/go-redis/redis"
)

// Adapter allows you to manipulate alarms in database
type Service struct {
	Adapter             *Adapter
	Cache               cache.Cache
	enableLastEventDate bool
}

// NewAdapter gives the correct alarm adapter. Give nil to the redis
// client and it will create a new redis.Client with the dedicated redis
// database for alarms.
func NewService(session *mgo.Session, redis *redis.Client, LastEventDate bool) Service {
	ma := NewAdapter(session)
	rc := NewCache(redis)
	tsfac := cache.NewThreadSafeCache(rc)

	return Service{
		Adapter:             &ma,
		Cache:               tsfac,
		enableLastEventDate: LastEventDate,
	}
}

// Insert a new alarm.
func (s *Service) Insert(alarm types.Alarm) error {
	err := s.Adapter.Insert(alarm)
	if err == nil {
		return s.Cache.Set(alarm)
	}
	return err
}

// Update updates an existing alarm or creates a new one in db.
func (s *Service) Update(alarm types.Alarm) error {
	err := s.Adapter.Update(alarm)
	if err == nil {
		return s.Cache.Set(alarm)
	}
	_, ok := err.(errt.NotFound)
	if ok {
		rerr := s.Cache.Drop(alarm.CacheID())
		if rerr != nil {
			return rerr
		}
	}
	return err
}

// Remove removes an existing alarm in db.
func (s *Service) Remove(alarm types.Alarm) error {
	err := s.Adapter.RemoveId(alarm.ID)
	if err == nil {
		return s.Cache.Drop(alarm.CacheID())
	}
	return err
}

// GetOneByEntityID find one opened alarm with his entity id.
// Note : a control is added to prevent fetching future alarms.
func (s *Service) GetOneByEntityID(connector, connectorName, id string) (types.Alarm, error) {
	alarmid := types.AlarmID(connector, connectorName, id)
	alarm, exists := s.Cache.Get(alarmid)
	if exists {
		return alarm.(types.Alarm), nil
	}

	al, err := s.Adapter.GetOpenedAlarm(connector, connectorName, id)

	if err != nil {
		rerr := s.Cache.Drop(alarmid)
		if rerr != nil {
			return al, rerr
		}
		return al, err
	}

	s.Cache.Set(al)
	return al, nil
}

// GetNotAckNotResolvedByEvent find all alarms which are not ack, not resolved and
// associated to an event. An alarm is associated to an event if his entities
// is link to the component of the given event.
func (s *Service) GetNotAckNotResolvedByEvent(event types.Event) ([]types.Alarm, error) {
	alarms, err := s.Adapter.GetNotAckNotResolvedAlarms(event.Component)

	var cacheErrors []error
	cacheErrors = make([]error, 0)
	for _, alarm := range alarms {
		error := s.Cache.Set(alarm)
		if error != nil {
			cacheErrors = append(cacheErrors, error)
		}
	}

	if len(cacheErrors) > 0 {
		return alarms, cacheErrors[len(cacheErrors)-1]
	}

	return alarms, err
}

// ResolveAlarms that have v.resolved to null
func (s *Service) ResolveAlarms(baggotTime time.Duration) (uint64, error) {
	alarms, err := s.Adapter.GetUnresolved()
	if err != nil {
		return 0, fmt.Errorf("unresolved alarms error: %v", err)
	}

	tnow := types.CpsTime{Time: time.Now()}

	closed := uint64(0)

	updatedAlarms := make([]types.Alarm, 0)
	for _, alarm := range alarms {
		if alarm.Closable(baggotTime) {
			alarm.Resolve(&tnow)
			updatedAlarms = append(updatedAlarms, alarm)
		}
	}

	err = s.Adapter.MassUpdate(updatedAlarms)

	if err != nil {
		return closed, fmt.Errorf("alarm resolution error: %v", err)
	}

	for _, alarm := range updatedAlarms {
		err = s.Cache.Drop(alarm.CacheID())
		if err != nil {
			return closed, fmt.Errorf("cache error: %v", err)
		}
	}

	return uint64(len(updatedAlarms)), nil
}

// ResolveSnoozes remove snooze state when snooze time has expired
func (s *Service) ResolveSnoozes() (uint64, error) {
	alarms, err := s.Adapter.GetAlarmsWithSnoozeMark()
	if err != nil {
		return 0, fmt.Errorf("snooze alarms error: %v", err)
	}

	snoozeResolved := uint64(0)
	updatedAlarms := make([]types.Alarm, 0)
	for _, alarm := range alarms {
		if !alarm.IsSnoozed() {
			alarm.UnSnooze()
			updatedAlarms = append(updatedAlarms, alarm)
		}
	}

	err = s.Adapter.MassUpdate(updatedAlarms)

	if err != nil {
		return snoozeResolved, fmt.Errorf("snooze resolution error: %v", err)
	}

	for _, alarm := range updatedAlarms {
		err = s.Cache.Drop(alarm.CacheID())
		if err != nil {
			return snoozeResolved, fmt.Errorf("cache error: %v", err)
		}
	}

	return uint64(len(updatedAlarms)), nil
}

// ResolveCancels close canceld alarms when time has expired
func (s *Service) ResolveCancels() (uint64, error) {
	alarms, err := s.Adapter.GetAlarmsWithCancelMark()
	if err != nil {
		return 0, fmt.Errorf("cancel alarms error: %v", err)
	}

	cancelResolved := uint64(0)
	updatedAlarms := make([]types.Alarm, 0)
	for _, alarm := range alarms {
		if !alarm.IsResolved() {
			delta := time.Since(alarm.Value.Canceled.Timestamp.Time)
			if int(delta.Seconds()) >= cps.CancelAutosolveDelay {
				tnow := types.CpsTime{Time: time.Now()}
				alarm.ResolveCancel(&tnow)
				updatedAlarms = append(updatedAlarms, alarm)
			}
		}
	}

	err = s.Adapter.MassUpdate(updatedAlarms)

	if err != nil {
		return cancelResolved, fmt.Errorf("cancel resolution error: %v", err)
	}

	for _, alarm := range updatedAlarms {
		err = s.Cache.Drop(alarm.AlarmID())
		if err != nil {
			return cancelResolved, fmt.Errorf("cache error: %v", err)
		}
	}

	return uint64(len(updatedAlarms)), nil
}

// ProcessAlarmEvent processes an event and creates/resolves alarms
func (s *Service) ProcessAlarmEvent(event *types.Event) error {
	event.Format()
	if event.Entity != nil {
		if !event.Entity.Enabled {
			return nil
		}
	}

	// Append the corresponding alarm if it exist already
	eid := event.GetEID()
	alarm, err := s.Adapter.GetOpenedAlarm(event.Connector, event.ConnectorName, eid)
	if err == nil {
		event.Alarm = &alarm
	}

	update, err := s.execTasks(event)
	if alarm.Value.DisplayName == "" {
		alarm.Value.DisplayName = types.GenDisplayName()
		update = true
	}

	if event.Alarm != nil && err == nil {
		if update || event.Alarm.CropSteps() {
			return s.Update(*event.Alarm)
		}
	}

	return err
}

// ProcessMassAck process a mass_ack event
func (s *Service) ProcessMassAck(event *types.Event, step types.AlarmStep) error {
	alarms, err := s.GetNotAckNotResolvedByEvent(*event)
	alarmsToUpdate := []types.Alarm{}

	if err != nil {
		return nil
	}

	for _, alarm := range alarms {
		if alarm.ID != event.Alarm.ID {
			step.Value = alarm.Value.State.Value
			alarm.Value.ACK = &step
			alarm.Value.Steps.Add(step)
			alarmsToUpdate = append(alarmsToUpdate, alarm)
		} else {
			alarmsToUpdate = append(alarmsToUpdate, alarm)
		}
	}

	for _, alarm := range alarmsToUpdate {
		err = s.Cache.Set(alarm)
	}

	s.Adapter.MassUpdate(alarmsToUpdate)

	return err
}

// Handle any event (temporary ; pass to python ?)
func (s *Service) execTasks(event *types.Event) (bool, error) {
	newStep := types.AlarmStep{
		Author:    event.Author,
		Message:   event.Output,
		Timestamp: types.CpsTime{Time: time.Now()},
	}

	switch event.EventType {
	case types.EventTypeWatcher:
		fallthrough
	case types.EventTypeCheck:
		return false, s.ProcessCheckOrWatcher(event)

	default:
		updateAlarm, newStep, err := s.execTasksRequiresAlarm(event, newStep)

		if err != nil {
			umeerr, ok := err.(errt.UnmanagedEventError)
			if ok {
				return false, umeerr
			}
			return false, err
		}

		if event.Alarm != nil {
			event.Alarm.Value.Steps.Add(newStep)

			if updateAlarm {
				event.Alarm.Update(*event)
			}
		}

		return true, nil
	}
}

// return updateAlarm, managedEventType, error
func (s *Service) execTasksRequiresAlarm(event *types.Event, newStep types.AlarmStep) (bool, types.AlarmStep, error) {
	updateAlarm := true

	if event.Alarm == nil {
		return false, newStep, nil
	}

	switch event.EventType {

	case types.EventTypeAck:
		if event.Alarm.IsAck() {
			return false, newStep, nil
		}

		newStep = event.Alarm.Ack(*event)
		var err error
		if event.MassAck == true {
			err = s.ProcessMassAck(event, newStep)
		}
		return updateAlarm, newStep, err

	case types.EventTypeAckremove:
		if event.Alarm.IsAck() {
			newStep = event.Alarm.Unack(*event)
		} else {
			updateAlarm = false
		}

	case types.EventTypeCancel:
		newStep = event.Alarm.Cancel(*event)

	case types.EventTypeUncancel:
		if event.Alarm.IsCanceled() {
			newStep = event.Alarm.Uncancel(*event)
		} else {
			updateAlarm = false
		}

	case types.EventTypeComment:
		newStep = event.Alarm.Comment(*event)

	case types.EventTypeDeclareTicket:
		newStep = event.Alarm.DeclareTicket(*event)

	case types.EventTypeAssocTicket:
		newStep = event.Alarm.AssocTicket(*event)

	case types.EventTypeChangestate:
		fallthrough
	case types.EventTypeKeepstate:
		newStep = event.Alarm.ChangeState(*event)

	case types.EventTypeSnooze:
		newStep = event.Alarm.Snooze(*event)

	case "linklist":
		log.Printf("unsupported event type linklist: %#v", event)
		//TODO mais je ne sais pas si ca passe par des events ca a confirmer

	case types.EventTypePBehavior:
		newStep.Type = types.AlarmStepPbehavior

	default:
		return false, newStep, errt.NewUnmanagedEventError(fmt.Errorf("event type: %v", event.EventType))
	}

	return updateAlarm, newStep, nil
}

// ProcessCheckOrWatcher ...
func (s *Service) ProcessCheckOrWatcher(event *types.Event) error {
	if event.Alarm == nil && event.State == types.AlarmStateOK {
		return nil
	}

	if event.Alarm == nil {
		newAlarm, err := types.NewAlarm(*event)

		if err == nil {
			event.Alarm = &newAlarm
			return s.Insert(*event.Alarm)
		}
	}

	updateAlarm := event.State != event.Alarm.Value.State.Value
	event.Alarm.Update(*event) // watcher events will be excluded here

	if s.enableLastEventDate {
		event.Alarm.Value.LastEventDate = event.Timestamp
		updateAlarm = true
	}

	if updateAlarm {
		return s.Update(*event.Alarm)
	}

	return nil
}
