package alarm_test

import (
	"fmt"
	"math/rand"
	"testing"
	"time"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/alarm"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/config"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/entity"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/event"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo"
	redis "git.canopsis.net/canopsis/go-revolution/lib/redis"
	. "github.com/smartystreets/goconvey/convey"
)

func TestNewAlarmDBAdapter(t *testing.T) {
	Convey("Setup", t, func() {
		db, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		defer db.Close()
		as := alarm.NewService(db, nil, false)
		So(as.Adapter.Collection.Name, ShouldEqual, alarm.AlarmCollectionName)
	})
}

func TestGetByEntityID(t *testing.T) {
	Convey("Setup", t, func() {
		db, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		defer db.Close()
		_, err = db.DB(cps.DbName).C(alarm.AlarmCollectionName).RemoveAll(nil)
		So(err, ShouldBeNil)

		as := alarm.NewService(db, nil, false)
		as.Cache.Flush()

		Convey("Add 2 alarm", func() {
			alarm, err := types.NewAlarm(getEventCritical())
			So(err, ShouldBeNil)
			So(as.Insert(alarm), ShouldBeNil)

			alarm.ID = types.NewAlarmID()
			alarm.EntityID = "cest/baleze"
			So(as.Insert(alarm), ShouldBeNil)

			alarm.ID = types.NewAlarmID()
			alarm.Value.State.Value = types.AlarmStateMajor
			So(as.Insert(alarm), ShouldBeNil)

			alarms, err := as.Adapter.GetByEntityID(alarm.EntityID)
			So(err, ShouldBeNil)
			So(len(alarms), ShouldEqual, 2)
			So(alarms[0].EntityID, ShouldEqual, "cest/baleze")
			So(alarms[0].Value.State.Value, ShouldEqual, types.AlarmStateCritical)
			So(alarms[1].EntityID, ShouldEqual, "cest/baleze")
			So(alarms[1].Value.State.Value, ShouldEqual, types.AlarmStateMajor)
		})
	})
}

func TestGetOneByEntityID(t *testing.T) {
	Convey("Setup", t, func() {
		db, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		defer db.Close()
		_, err = db.DB(cps.DbName).C(alarm.AlarmCollectionName).RemoveAll(nil)
		So(err, ShouldBeNil)

		r, err := redis.NewSession(redis.CacheAlarm)
		So(err, ShouldBeNil)
		defer r.Close()

		r.FlushDB()

		as := alarm.NewService(db, nil, false)
		as.Cache.Flush()

		Convey("Insert alarm", func() {
			evt := getEventCritical()
			nalarm, err := types.NewAlarm(evt)
			So(err, ShouldBeNil)
			So(as.Insert(nalarm), ShouldBeNil)

			dalarm, err := as.GetOneByEntityID(evt.Connector, evt.ConnectorName, nalarm.EntityID)
			So(err, ShouldBeNil)
			So(dalarm.EntityID, ShouldEqual, nalarm.EntityID)
			So(dalarm.Value.State.Value, ShouldEqual, nalarm.Value.State.Value)
		})
	})
}

// GenerateRandomEvent needs a call to rand.Seed(int) before you use it.
func GenerateRandomEvent(ncomponents int, nresources int, critprob int64) types.Event {
	tsnow := types.CpsTime{Time: time.Now()}
	event := types.Event{
		Connector:     "feeder",
		ConnectorName: "feeder",
		SourceType:    "resource",
		EventType:     "check",
		Component:     fmt.Sprintf("component_%d", rand.Intn(ncomponents)),
		Resource:      fmt.Sprintf("resource_%d", rand.Intn(nresources)),
		State:         0,
		Timestamp:     tsnow,
	}
	statep := rand.Int63n(100)
	if statep > 100-critprob {
		event.State = 3
	}
	return event
}

func BenchmarkProcessAlarmEvent(b *testing.B) {
	rand.Seed(13374219847211729)
	session, err := mongo.NewSession(mongo.Timeout)
	if err != nil {
		b.Fatal(err)
	}
	defer session.Close()

	as := alarm.NewService(session, nil, false)
	edba := entity.NewAdapter(session, nil)
	evtdba := event.NewEventDBAdapter(&as, &edba)

	as.Cache.Flush()
	edba.Cache.Flush()

	ncomp := 100
	nres := 10
	nnn := ncomp * nres
	events := make([]types.Event, nnn)
	for i := 0; i < nnn; i++ {
		events[i] = GenerateRandomEvent(ncomp, nres, 20)
	}

	session.DB(cps.DbName).C(alarm.AlarmCollectionName).RemoveAll(nil)
	evtdba.AlarmService.Cache.Flush()

	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		evt := events[i%nnn]
		evt, report := evtdba.Enrich(evt)
		if report.Error != nil {
			b.Fatal(report.Error)
		}

		err = evtdba.AlarmService.ProcessAlarmEvent(&evt)
		if err != nil {
			b.Fatal(err)
		}
	}
}

func getEventOK() types.Event {
	now := types.CpsTime{Time: time.Now()}
	return types.Event{
		Component:     "spartatouille",
		Resource:      "xerces",
		EventType:     types.EventTypeCheck,
		SourceType:    types.SourceTypeResource,
		Connector:     "convey",
		ConnectorName: "convey",
		State:         types.AlarmStateOK,
		Output:        "Respect and honor",
		Timestamp:     now,
	}
}

func getEventMajor() types.Event {
	now := types.CpsTime{Time: time.Now()}
	return types.Event{
		Component:     "spartatouille",
		Resource:      "xerces",
		EventType:     types.EventTypeCheck,
		SourceType:    types.SourceTypeResource,
		Connector:     "convey",
		ConnectorName: "convey",
		State:         types.AlarmStateMajor,
		Output:        "Earn these shields, boys!",
		Timestamp:     now,
	}
}

func getEventCritical() types.Event {
	now := types.CpsTime{Time: time.Now()}
	return types.Event{
		Component:     "spartatouille",
		Resource:      "xerces",
		EventType:     types.EventTypeCheck,
		SourceType:    types.SourceTypeResource,
		Connector:     "convey",
		ConnectorName: "convey",
		State:         types.AlarmStateCritical,
		Output:        "Spartans !",
		Timestamp:     now,
	}
}

func getEventAck() types.Event {
	now := types.CpsTime{Time: time.Now()}
	return types.Event{
		Component:     "spartatouille",
		Resource:      "xerces",
		EventType:     "ack",
		SourceType:    types.SourceTypeResource,
		Connector:     "convey",
		ConnectorName: "convey",
		State:         types.AlarmStateCritical, // this state must be critical to ensure we don't care about a state change while ack
		Output:        "It's an honor to die at your side",
		Timestamp:     now,
	}
}

func getEventAckRemove() types.Event {
	now := types.CpsTime{Time: time.Now()}
	return types.Event{
		Component:     "spartatouille",
		Resource:      "xerces",
		EventType:     "ackremove",
		SourceType:    types.SourceTypeResource,
		Connector:     "convey",
		ConnectorName: "convey",
		State:         types.AlarmStateMajor, // this state must be major to ensure we don't care about a state change while ackremove
		Output:        "It's an honor to have lived at yours",
		Timestamp:     now,
	}
}

func getEventSnooze() types.Event {
	now := types.CpsTime{Time: time.Now()}
	duration := types.CpsNumber(7 * 24 * 60 * 60)
	return types.Event{
		Component:     "spartatouille",
		Resource:      "xerces",
		EventType:     types.EventTypeSnooze,
		Connector:     "convey",
		ConnectorName: "convey",
		Output:        "Traitoro is a traitor",
		Timestamp:     now,
		Duration:      &duration,
	}
}

func getEventCancel() types.Event {
	now := types.CpsTime{Time: time.Now()}
	return types.Event{
		Component:     "spartatouille",
		Resource:      "xerces",
		EventType:     types.EventTypeCancel,
		Connector:     "convey",
		ConnectorName: "convey",
		Output:        "King Xerxès is a king",
		Timestamp:     now,
	}
}

func getEventUncancel() types.Event {
	now := types.CpsTime{Time: time.Now()}
	return types.Event{
		Component:     "spartatouille",
		Resource:      "xerces",
		EventType:     types.EventTypeUncancel,
		Connector:     "convey",
		ConnectorName: "convey",
		Output:        "King Xerxès is not a winner",
		Timestamp:     now,
	}
}

func getEventChangeState() types.Event {
	now := types.CpsTime{Time: time.Now()}
	return types.Event{
		Component:     "spartatouille",
		Resource:      "xerces",
		EventType:     types.EventTypeChangestate,
		State:         3,
		Connector:     "convey",
		ConnectorName: "convey",
		Output:        "Léonidas says that there is not turning back",
		Timestamp:     now,
	}
}

func TestAlarmDBAdapter(t *testing.T) {
	Convey("Setup", t, func() {
		mongo, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		So(mongo, ShouldNotBeNil)
		defer mongo.Close()

		redis, err := redis.NewSession(redis.CacheAlarm)
		So(err, ShouldBeNil)
		defer redis.Close()

		as := alarm.NewService(mongo, nil, false)
		edba := entity.NewAdapter(mongo, nil)
		evtdba := event.NewEventDBAdapter(&as, &edba)
		So(evtdba.AlarmService, ShouldNotBeNil)
		So(evtdba.EntityAdapter, ShouldNotBeNil)

		as.Cache.Flush()
		edba.Cache.Flush()

		mongo.DB(cps.DbName).C(entity.EntityCollectionName).RemoveAll(nil)
		mongo.DB(cps.DbName).C(alarm.AlarmCollectionName).RemoveAll(nil)

		Convey("Given an Ack event on an empty alarm", func() {
			evt := getEventAck()

			Convey("Then it should not crash", func() {
				err = evtdba.AlarmService.ProcessAlarmEvent(&evt)

				So(err, ShouldBeNil)
				So(evt.Alarm, ShouldBeNil)
			})
		})

		Convey("Given an OK event", func() {
			evt := getEventOK()
			evt, report := evtdba.Enrich(evt)

			So(report.Error, ShouldBeNil)
			So(report.HasEntity, ShouldBeFalse)

			err = evtdba.AlarmService.ProcessAlarmEvent(&evt)

			So(err, ShouldBeNil)
			So(evt.Alarm, ShouldBeNil)

			Convey("Then first MAJOR event", func() {
				evt := getEventMajor()
				evt, report := evtdba.Enrich(evt)

				So(report.Error, ShouldBeNil)
				So(report.HasEntity, ShouldBeFalse)

				err = evtdba.AlarmService.ProcessAlarmEvent(&evt)

				So(err, ShouldBeNil)
				So(evt.Alarm, ShouldNotBeNil)
				So(len(evt.Alarm.Value.Steps), ShouldEqual, 2) // stateinc and statusinc
				So(evt.Alarm.Value.State.Value, ShouldEqual, types.AlarmStateMajor)
				So(evt.Alarm.Value.State.Timestamp.Unix(), ShouldEqual, evt.Timestamp.Unix())

				Convey("Then second MAJOR event", func() {
					evt := getEventMajor()
					evt, report := evtdba.Enrich(evt)

					So(report.Error, ShouldBeNil)
					So(evt, ShouldNotBeNil)

					err = evtdba.AlarmService.ProcessAlarmEvent(&evt)

					So(err, ShouldBeNil)
					So(evt.Alarm, ShouldNotBeNil)
					So(len(evt.Alarm.Value.Steps), ShouldEqual, 2)
					So(evt.Alarm.Value.State.Value, ShouldEqual, types.AlarmStateMajor)

					Convey("Then ACK event", func() {
						evt := getEventAck()
						evt, report := evtdba.Enrich(evt)

						So(report.Error, ShouldBeNil)
						So(report.HasEntity, ShouldBeFalse)
						So(evt, ShouldNotBeNil)

						err = evtdba.AlarmService.ProcessAlarmEvent(&evt)

						So(err, ShouldBeNil)
						So(len(evt.Alarm.Value.Steps), ShouldEqual, 3)
						lastStep, err := evt.Alarm.Value.Steps.Last()
						So(err, ShouldBeNil)
						So(lastStep.Type, ShouldEqual, types.AlarmStepAck)
						So(evt.Alarm.IsAck(), ShouldBeTrue)

						Convey("Then CRITICAL event", func() {
							evt := getEventCritical()
							evt, report := evtdba.Enrich(evt)

							So(report.Error, ShouldBeNil)
							So(report.HasEntity, ShouldBeFalse)
							So(evt, ShouldNotBeNil)

							err = evtdba.AlarmService.ProcessAlarmEvent(&evt)

							So(err, ShouldBeNil)
							So(len(evt.Alarm.Value.Steps), ShouldEqual, 4)
							So(evt.Alarm.Value.State.Value, ShouldEqual, types.AlarmStateCritical)
							lastStep, err = evt.Alarm.Value.Steps.Last()
							So(err, ShouldBeNil)
							So(lastStep.Type, ShouldEqual, types.AlarmStepStateIncrease)
							So(lastStep.Value, ShouldEqual, types.AlarmStateCritical)

							Convey("Then ACKREMOVE event", func() {
								evt := getEventAckRemove()
								evt, report := evtdba.Enrich(evt)

								So(report.Error, ShouldBeNil)
								So(report.HasEntity, ShouldBeFalse)
								So(evt, ShouldNotBeNil)

								err = evtdba.AlarmService.ProcessAlarmEvent(&evt)

								So(err, ShouldBeNil)
								So(len(evt.Alarm.Value.Steps), ShouldEqual, 5)
								lastStep, err = evt.Alarm.Value.Steps.Last()
								So(err, ShouldBeNil)
								So(lastStep.Type, ShouldEqual, types.AlarmStepAckRemove)
								So(evt.Alarm.IsAck(), ShouldBeFalse)

								Convey("Then SNOOZE event", func() {
									evt := getEventSnooze()
									evt, report := evtdba.Enrich(evt)

									So(report.Error, ShouldBeNil)
									So(report.HasEntity, ShouldBeFalse)
									So(evt, ShouldNotBeNil)

									err := evtdba.AlarmService.ProcessAlarmEvent(&evt)
									So(err, ShouldBeNil)
									So(len(evt.Alarm.Value.Steps), ShouldEqual, 6)
									So(evt.Alarm.IsSnoozed(), ShouldBeTrue)

									Convey("Then Cancel event", func() {
										evt := getEventCancel()
										evt, report := evtdba.Enrich(evt)

										So(report.Error, ShouldBeNil)
										So(report.HasEntity, ShouldBeFalse)
										So(evt, ShouldNotBeNil)

										err := evtdba.AlarmService.ProcessAlarmEvent(&evt)
										So(err, ShouldBeNil)
										So(len(evt.Alarm.Value.Steps), ShouldEqual, 8) // cancel + stateinc
										So(evt.Alarm.IsCanceled(), ShouldBeTrue)

										Convey("Then Uncancel event", func() {
											evt := getEventUncancel()
											evt, report := evtdba.Enrich(evt)

											So(report.Error, ShouldBeNil)
											So(report.HasEntity, ShouldBeFalse)
											So(evt, ShouldNotBeNil)

											err := evtdba.AlarmService.ProcessAlarmEvent(&evt)

											So(err, ShouldBeNil)
											So(len(evt.Alarm.Value.Steps), ShouldEqual, 10) // uncancel + statedec
											So(evt.Alarm.IsCanceled(), ShouldBeFalse)

											Convey("Then changestate event", func() {
												evt := getEventChangeState()
												evt, report := evtdba.Enrich(evt)

												So(report.Error, ShouldBeNil)
												So(report.HasEntity, ShouldBeFalse)
												So(evt, ShouldNotBeNil)

												err := evtdba.AlarmService.ProcessAlarmEvent(&evt)

												So(err, ShouldBeNil)
												So(len(evt.Alarm.Value.Steps), ShouldEqual, 11)
												So(evt.Alarm.IsStateLocked(), ShouldBeTrue)
												So(evt.Alarm.Value.State.Value, ShouldEqual, evt.State)
											})
										})
									})
								})
							})
						})
					})
				})
			})
		})
	})
}

func TestResolveAlarms(t *testing.T) {
	Convey("Setup", t, func() {
		config.ResetConf()
		mongo, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		So(mongo, ShouldNotBeNil)
		defer mongo.Close()

		as := alarm.NewService(mongo, nil, false)
		edba := entity.NewAdapter(mongo, nil)
		evtdba := event.NewEventDBAdapter(&as, &edba)
		So(evtdba.AlarmService, ShouldNotBeNil)
		So(evtdba.EntityAdapter, ShouldNotBeNil)

		_, err = mongo.DB(cps.DbName).C(alarm.AlarmCollectionName).RemoveAll(nil)
		So(err, ShouldBeNil)
		r := evtdba.AlarmService.Cache.Flush()
		So(r, ShouldBeNil)
		r = evtdba.EntityAdapter.Cache.Flush()
		So(r, ShouldBeNil)

		Convey("Create an alarm and process it", func() {
			evt := getEventCritical()
			evt, report := evtdba.Enrich(evt)

			So(report.Error, ShouldBeNil)
			err = evtdba.AlarmService.ProcessAlarmEvent(&evt)
			So(err, ShouldBeNil)

			Convey("Get unresolved alarms", func() {
				alarms, err := evtdba.AlarmService.Adapter.GetUnresolved()
				So(err, ShouldBeNil)
				So(len(alarms), ShouldEqual, 1)
				So(alarms[0].Closable(time.Second), ShouldBeFalse)

				Convey("Run ResolveAlarms", func() {
					resolvecount, err := evtdba.AlarmService.ResolveAlarms(time.Second * 1)
					So(err, ShouldBeNil)
					So(resolvecount, ShouldEqual, 0)

					alarms, err = evtdba.AlarmService.Adapter.GetUnresolved()
					So(len(alarms), ShouldEqual, 1)

					Convey("Send OK event", func() {
						evt := getEventOK()
						evt, report := evtdba.Enrich(evt)
						So(report.Error, ShouldBeNil)
						So(report.HasEntity, ShouldBeFalse)

						err = evtdba.AlarmService.ProcessAlarmEvent(&evt)
						So(err, ShouldBeNil)

						alarms, _ := evtdba.AlarmService.Adapter.GetUnresolved()
						So(len(alarms), ShouldEqual, 1)

						Convey("Run ResolveAlarms again after waiting baggot time", func() {
							time.Sleep(time.Second * 1)
							resolvecount, err := evtdba.AlarmService.ResolveAlarms(time.Second * 1)
							So(err, ShouldBeNil)
							So(resolvecount, ShouldEqual, 1)

							alarm, exists := evtdba.AlarmService.Cache.Get(evt.GetEID())
							So(alarm, ShouldBeNil)
							So(exists, ShouldBeFalse)

							alarms, err = evtdba.AlarmService.Adapter.GetUnresolved()
							So(len(alarms), ShouldEqual, 0)

							Convey("Send Critical event", func() {
								evt := getEventCritical()
								evt, report := evtdba.Enrich(evt)
								So(report.Error, ShouldBeNil)

								err = evtdba.AlarmService.ProcessAlarmEvent(&evt)
								So(err, ShouldBeNil)

								alarms, _ := evtdba.AlarmService.Adapter.GetUnresolved()
								So(len(alarms), ShouldEqual, 1)
							})
						})
					})
				})
			})
		})
	})
}

func TestResolveSnoozes(t *testing.T) {
	Convey("Setup alarm adapter", t, func() {
		config.ResetConf()
		mongo, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		So(mongo, ShouldNotBeNil)
		defer mongo.Close()

		as := alarm.NewService(mongo, nil, false)
		edba := entity.NewAdapter(mongo, nil)
		evtdba := event.NewEventDBAdapter(&as, &edba)
		So(evtdba.AlarmService, ShouldNotBeNil)
		So(evtdba.EntityAdapter, ShouldNotBeNil)

		_, err = mongo.DB(cps.DbName).C(alarm.AlarmCollectionName).RemoveAll(nil)
		So(err, ShouldBeNil)
		r := evtdba.AlarmService.Cache.Flush()
		So(r, ShouldBeNil)
		r = evtdba.EntityAdapter.Cache.Flush()
		So(r, ShouldBeNil)

		Convey("Create an alarm and process it", func() {
			evt := getEventCritical()
			evt, report := evtdba.Enrich(evt)
			So(report.Error, ShouldBeNil)
			So(report.HasEntity, ShouldBeFalse)
			err := evtdba.AlarmService.ProcessAlarmEvent(&evt)
			So(err, ShouldBeNil)

			Convey("Then there is no snoozed alarm", func() {
				resolvecount, err := evtdba.AlarmService.ResolveSnoozes()
				So(err, ShouldBeNil)
				So(resolvecount, ShouldEqual, 0)

				alarms, err := evtdba.AlarmService.Adapter.GetAlarmsWithSnoozeMark()
				So(len(alarms), ShouldEqual, 0)
			})

			Convey("By snoozing an alarm", func() {
				evt := getEventSnooze()
				evt, report := evtdba.Enrich(evt)
				So(report.Error, ShouldBeNil)
				So(report.HasEntity, ShouldBeFalse)

				err := evtdba.AlarmService.ProcessAlarmEvent(&evt)
				So(err, ShouldBeNil)

				Convey("Then ResolveSnoozes() should not resolve it", func() {
					resolvecount, err := evtdba.AlarmService.ResolveSnoozes()
					So(err, ShouldBeNil)
					So(resolvecount, ShouldEqual, 0)

					alarms, err := evtdba.AlarmService.Adapter.GetAlarmsWithSnoozeMark()
					So(err, ShouldBeNil)
					So(len(alarms), ShouldEqual, 1)
				})
			})

			Convey("By past snoozing an alarm", func() {
				evt := getEventSnooze()
				in_past := types.CpsNumber(-100)
				evt.Duration = &in_past
				evt, report := evtdba.Enrich(evt)
				So(report.Error, ShouldBeNil)
				So(report.HasEntity, ShouldBeFalse)

				err := evtdba.AlarmService.ProcessAlarmEvent(&evt)
				So(err, ShouldBeNil)

				Convey("Then ResolveSnoozes() should resolve it", func() {
					resolvecount, err := evtdba.AlarmService.ResolveSnoozes()
					So(err, ShouldBeNil)
					So(resolvecount, ShouldEqual, 1)

					alarms, err := evtdba.AlarmService.Adapter.GetAlarmsWithSnoozeMark()
					So(err, ShouldBeNil)
					So(len(alarms), ShouldEqual, 0)
				})
			})
		})
	})
}

func TestResolveCancels(t *testing.T) {
	Convey("Setup alarm adapter", t, func() {
		config.ResetConf()
		mongo, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		So(mongo, ShouldNotBeNil)
		defer mongo.Close()

		as := alarm.NewService(mongo, nil, false)
		edba := entity.NewAdapter(mongo, nil)
		evtdba := event.NewEventDBAdapter(&as, &edba)
		So(evtdba.AlarmService, ShouldNotBeNil)
		So(evtdba.EntityAdapter, ShouldNotBeNil)

		_, err = mongo.DB(cps.DbName).C(alarm.AlarmCollectionName).RemoveAll(nil)
		So(err, ShouldBeNil)
		r := evtdba.AlarmService.Cache.Flush()
		So(r, ShouldBeNil)
		r = evtdba.EntityAdapter.Cache.Flush()
		So(r, ShouldBeNil)

		Convey("Create an alarm and process it", func() {
			evt := getEventCritical()
			evt, report := evtdba.Enrich(evt)
			So(report.Error, ShouldBeNil)
			So(report.HasEntity, ShouldBeFalse)
			err := evtdba.AlarmService.ProcessAlarmEvent(&evt)
			So(err, ShouldBeNil)

			Convey("Then there is no canceled alarm", func() {
				resolvecount, err := evtdba.AlarmService.ResolveCancels()
				So(err, ShouldBeNil)
				So(resolvecount, ShouldEqual, 0)

				alarms, err := evtdba.AlarmService.Adapter.GetAlarmsWithCancelMark()
				So(len(alarms), ShouldEqual, 0)
			})

			Convey("By canceling an alarm", func() {
				evt := getEventCancel()
				evt, report := evtdba.Enrich(evt)
				So(report.Error, ShouldBeNil)
				So(report.HasEntity, ShouldBeFalse)

				err := evtdba.AlarmService.ProcessAlarmEvent(&evt)
				So(err, ShouldBeNil)

				Convey("Then ResolveCancels() should not resolve it", func() {
					resolvecount, err := evtdba.AlarmService.ResolveCancels()
					So(err, ShouldBeNil)
					So(resolvecount, ShouldEqual, 0)

					alarms, err := evtdba.AlarmService.Adapter.GetAlarmsWithCancelMark()
					So(err, ShouldBeNil)
					So(len(alarms), ShouldEqual, 1)

					Convey("By moving cancel action in the past", func() {
						in_past := types.CpsTime{
							Time: time.Now().Add(cps.CancelAutosolveDelay * -2 * time.Second),
						}
						alarm := alarms[0]
						alarm.Value.Canceled.Timestamp = in_past
						err = as.Update(alarm)
						So(err, ShouldBeNil)

						Convey("Then ResolveCancels() should resolve it", func() {
							resolvecount, err := evtdba.AlarmService.ResolveCancels()
							So(err, ShouldBeNil)
							So(resolvecount, ShouldEqual, 1)

							alarms, err := evtdba.AlarmService.Adapter.GetAlarmsWithCancelMark()
							So(err, ShouldBeNil)
							So(len(alarms), ShouldEqual, 1)
						})
					})
				})
			})
		})
	})
}

func TestProcessCheckOrWatcher(t *testing.T) {
	Convey("Setup alarm adapter", t, func() {
		config.ResetConf()
		mongo, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		So(mongo, ShouldNotBeNil)
		defer mongo.Close()

		as := alarm.NewService(mongo, nil, false)
		edba := entity.NewAdapter(mongo, nil)
		evtdba := event.NewEventDBAdapter(&as, &edba)
		So(evtdba.AlarmService, ShouldNotBeNil)
		So(evtdba.EntityAdapter, ShouldNotBeNil)

		mongo.DB(cps.DbName).C(entity.EntityCollectionName).RemoveAll(nil)
		_, err = mongo.DB(cps.DbName).C(alarm.AlarmCollectionName).RemoveAll(nil)
		So(err, ShouldBeNil)
		r := evtdba.AlarmService.Cache.Flush()
		So(r, ShouldBeNil)
		r = evtdba.EntityAdapter.Cache.Flush()
		So(r, ShouldBeNil)

		Convey("Create a check event and process it", func() {
			evt := getEventMajor()
			evt, report := evtdba.Enrich(evt)
			So(report.Error, ShouldBeNil)
			err := as.ProcessCheckOrWatcher(&evt)
			So(err, ShouldBeNil)

			evt = getEventCritical()
			err = as.ProcessAlarmEvent(&evt)
			So(err, ShouldBeNil)
			So(evt.Alarm, ShouldNotBeNil)
			So(evt.Alarm.Value.Steps, ShouldNotBeNil)
			So(len(evt.Alarm.Value.Steps), ShouldEqual, 3) // state inc * 2, status inc

			err = as.ProcessCheckOrWatcher(&evt)
			So(err, ShouldBeNil)

			So(evt.State, ShouldEqual, evt.Alarm.Value.State.Value)
			So(evt.State, ShouldEqual, types.AlarmStateCritical)
		})
	})
}
