package config

import (
	"errors"
	"fmt"
	"os"
	"time"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/heartbeat"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo"
	"git.canopsis.net/canopsis/go-revolution/lib/utils"
	"github.com/BurntSushi/toml"
	mgo "github.com/globalsign/mgo"
)

// Some default values related to configuration
const (
	ConfigCollectionName = "configuration"
	ConfigFileName       = "default_configuration.toml"
	ConfigKeyName        = "global_config"
	EnvCpsDefaultCfgFile = "CPS_DEFAULT_CFG"
)

// ErrNotFound when no configuration found anywhere
var ErrNotFound = errors.New("conf not found")

var globalConfigDBAdapter *Adapter
var globalConf *Conf

// SectionAlarm ...
type SectionAlarm struct {
	FlappingFreqLimit   int    `toml:"FlappingFreqLimit"`
	FlappingInterval    int    `toml:"FlappingInterval"`
	StealthyInterval    int    `toml:"StealthyInterval"`
	BaggotTime          string `toml:"BaggotTime"`
	EnableLastEventDate bool   `toml:"EnableLastEventDate"`
}

// SectionGlobal ...
type SectionGlobal struct {
	PrefetchCount int `toml:"PrefetchCount"`
	PrefetchSize  int `toml:"PrefetchSize"`
}

// ConfHeartBeatItem contains the required informations for the HeartBeat engine
// to work.
type ConfHeartBeatItem struct {
	Mappings    []map[string]string `toml:"mappings"`
	MaxDuration string              `toml:"maxduration"`
	Registered  map[string]int      `toml:"registered"`
}

// ConfResetGlobal sets global conf state to nil
func ConfResetGlobal() {
	globalConf = nil
	globalConfigDBAdapter = nil
}

// ToHeartBeatItem converts a configuration struct to the actual HeartBeatItem struct.
func (cli ConfHeartBeatItem) ToHeartBeatItem() (heartbeat.Item, error) {
	maxDuration, err := time.ParseDuration(cli.MaxDuration)
	if err != nil {
		return heartbeat.NewItem(0), err
	}
	li := heartbeat.NewItem(maxDuration)
	for _, mapping := range cli.Mappings {
		for field, matches := range mapping {
			li.AddMapping(field, matches)
		}
	}
	return li, nil
}

// SectionHeartBeat ...
type SectionHeartBeat struct {
	Items []ConfHeartBeatItem `toml:"items"`
}

// Conf represents a generic configuration object.
type Conf struct {
	ID        string           `bson:"_id,omitempty" toml:"omitempty"`
	Global    SectionGlobal    `bson:"global" toml:"global"`
	Alarm     SectionAlarm     `bson:"alarm" toml:"alarm"`
	HeartBeat SectionHeartBeat `bson:"heartbeat" toml:"heartbeat"`
}

// Adapter to the concrete database
type Adapter struct {
	Collection *mgo.Collection
}

// NewConfigDBAdapter give the correct config adapter
func NewConfigDBAdapter(session *mgo.Session) Adapter {
	return Adapter{Collection: session.DB(cps.DbName).C(ConfigCollectionName)}
}

// ResetConf reinitialize all config objects and calls GetConf()
func ResetConf() {
	globalConf = nil
	globalConfigDBAdapter = nil
	GetConf()
}

// CleanupConf can be called after a successful GetConf() to remove remote configuration.
func CleanupConf() error {
	if globalConfigDBAdapter != nil {
		return globalConfigDBAdapter.CleanupConf()
	}
	return nil
}

// RefreshConf updates the local configuration from the live database.
func RefreshConf() *Conf {
	GetConf()
	conf, err := globalConfigDBAdapter.ReadWholeConf()
	utils.ExitOnError(err, "conf refresh error")
	globalConf = &conf
	return GetConf()
}

// GetConf instanciate and return a global Conf object and adapter
func GetConf() *Conf {
	if globalConfigDBAdapter == nil {
		session, err := mongo.NewSession(mongo.Timeout)
		// FIXIT: avoid panic
		utils.FailOnError(err, "mongo session error")
		dbAdapter := NewConfigDBAdapter(session)
		globalConfigDBAdapter = &dbAdapter
	}
	if globalConf == nil {
		conf, err := globalConfigDBAdapter.ReadWholeConf()
		utils.FailOnError(err, "conf read error")
		globalConf = &conf
	}
	return globalConf
}

// Save push a modified Conf into db
func (c Conf) Save() *mgo.ChangeInfo {
	changeInfo, _ := globalConfigDBAdapter.UpsertConfig(c)
	globalConf = nil
	return changeInfo
}

// CleanupConf remove the conf document in mongo.
func (c *Adapter) CleanupConf() error {
	return c.Collection.RemoveId(ConfigKeyName)
}

// GetByID find a document from an id.
func (c *Adapter) GetByID(id string) (Conf, error) {
	conf := Conf{}
	err := c.Collection.FindId(id).One(&conf)
	if err == mgo.ErrNotFound {
		return conf, ErrNotFound
	}
	return conf, err
}

// ReadWholeConf return the whole config document
func (c *Adapter) ReadWholeConf() (Conf, error) {
	conf, err := c.GetByID(ConfigKeyName)
	if err == ErrNotFound {
		// Insert default config values
		cfgfile := os.Getenv(EnvCpsDefaultCfgFile)
		if cfgfile == "" {
			cfgfile = ConfigFileName
		}
		source, err := utils.ReadFile(cfgfile)
		if err != nil {
			return Conf{}, fmt.Errorf("conf read whole conf: %v", err)
		}
		newConf, err := ReadTomlConf(source)
		if err != nil {
			return newConf, fmt.Errorf("conf read whole conf: %v", err)
		}
		newConf.Save()
		return newConf, nil
	}
	return conf, err
}

// UpsertConfig upsert a config in mongo.
func (c *Adapter) UpsertConfig(conf Conf) (*mgo.ChangeInfo, error) {
	conf.ID = ConfigKeyName
	changeInfo, err := c.Collection.UpsertId(ConfigKeyName, conf)

	return changeInfo, err
}

// ReadTomlConf convert toml configuration to Conf struct
func ReadTomlConf(source []byte) (Conf, error) {
	var conf Conf
	err := toml.Unmarshal(source, &conf)
	if err != nil {
		return conf, fmt.Errorf("cannot parse toml: %v", err)
	}
	return conf, nil
}
