package config

import (
	"io/ioutil"
	"os"
	"testing"

	"git.canopsis.net/canopsis/go-revolution/lib/mongo"
	. "github.com/smartystreets/goconvey/convey"
)

func GetTestConf() Conf {
	return Conf{
		Global: SectionGlobal{
			PrefetchCount: 10000,
			PrefetchSize:  0,
		},
		Alarm: SectionAlarm{
			FlappingFreqLimit: 1,
			FlappingInterval:  1,
			StealthyInterval:  100,
		},
	}
}

func TestConfWriteAndRead(t *testing.T) {
	Convey("Given a session, config DB adapter and a config", t, func() {
		s, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		defer s.Close()
		c := NewConfigDBAdapter(s)
		_ = GetConf()

		source := GetTestConf()

		Convey("Init should work", func() {
			info, err := c.UpsertConfig(source)
			So(err, ShouldBeNil)
			So(info.Matched, ShouldEqual, 1)
		})

		Convey("Readed conf should be good", func() {
			conf, err := c.ReadWholeConf()
			So(err, ShouldBeNil)
			So(conf.Alarm.FlappingFreqLimit, ShouldEqual, 1)
		})
	})
}

func TestConfSave(t *testing.T) {
	Convey("Given a session, config DB adapter and a config", t, func() {
		s, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		defer s.Close()
		c := NewConfigDBAdapter(s)
		source := GetConf()

		Convey("When we set a value", func() {
			source.Alarm.FlappingInterval = 666
			info := source.Save()
			So(info.Matched, ShouldEqual, 1)

			Convey("The the value is on the database", func() {
				conf, err := c.ReadWholeConf()
				So(err, ShouldBeNil)
				So(conf.Alarm.FlappingInterval, ShouldEqual, 666)
			})

			c.CleanupConf()
		})
	})
}

func TestConfToHeartBeat(t *testing.T) {
	Convey("Given a ConfHeartBeatItem", t, func() {
		cli := ConfHeartBeatItem{
			Mappings:    make([]map[string]string, 0),
			MaxDuration: "10s",
			Registered:  make(map[string]int),
		}

		cli.Mappings = append(cli.Mappings, map[string]string{"connector": "test"})

		Convey("I can translate to a HeartBeatItem", func() {
			li, err := cli.ToHeartBeatItem()
			So(err, ShouldBeNil)
			So(li.Registered["connector"], ShouldEqual, 0)
			So(li.Mappings[0]["connector"], ShouldEqual, "test")
		})

		Convey("I cannot translate a bad MaxDuration value", func() {
			cli.MaxDuration = "war string"
			_, err := cli.ToHeartBeatItem()
			So(err, ShouldNotBeNil)
		})
	})
}

func TestGetConfWrongMongo(t *testing.T) {
	Convey("Given a bad mongo url", t, func() {
		ConfResetGlobal()

		ou := os.Getenv(mongo.EnvURL)
		os.Setenv(mongo.EnvURL, "howmanytimeshaveitoldyaidontexist?")
		So(func() { GetConf() }, ShouldPanic)
		os.Setenv(mongo.EnvURL, ou)

		ConfResetGlobal()
	})
}

func TestGetConfID(t *testing.T) {
	Convey("Given a bad conf id", t, func() {
		id := "hollycowimustonceagainsayidontexist"
		So(func() { GetConf() }, ShouldNotPanic)
		_, err := globalConfigDBAdapter.GetByID(id)
		So(err, ShouldEqual, ErrNotFound)
	})
}

func TestConfigCfg(t *testing.T) {
	oe := os.Getenv(EnvCpsDefaultCfgFile)
	Convey("Given an empty env var", t, func() {
		So(func() { ResetConf() }, ShouldNotPanic)
		So(CleanupConf(), ShouldBeNil)
		os.Setenv(EnvCpsDefaultCfgFile, "")
		So(func() { ResetConf() }, ShouldPanic)
	})
	os.Setenv(EnvCpsDefaultCfgFile, oe)

	Convey("Given a bad default config path", t, func() {
		So(func() { ResetConf() }, ShouldNotPanic)
		So(CleanupConf(), ShouldBeNil)
		os.Setenv(EnvCpsDefaultCfgFile, "woooowyouaskreallytooooooMUUUCH")
		So(func() { ResetConf() }, ShouldPanic)
	})
	os.Setenv(EnvCpsDefaultCfgFile, oe)

	Convey("Given a bad toml conf file", t, func() {
		f, err := ioutil.TempFile("", "test-canopsis-go")
		So(err, ShouldBeNil)
		defer os.Remove(f.Name())

		_, err = f.WriteString("pwahaha[[]]]]")
		So(err, ShouldBeNil)

		Convey("When i try to parse, i fail", func() {
			So(func() { ResetConf() }, ShouldNotPanic)
			So(CleanupConf(), ShouldBeNil)
			os.Setenv(EnvCpsDefaultCfgFile, f.Name())
			So(func() { ResetConf() }, ShouldPanic)
		})
		os.Setenv(EnvCpsDefaultCfgFile, oe)
	})
}
