package context

import (
	"log"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/entity"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
)

type builder struct {
	pendingOk     map[string]EntityState
	pending       map[string]EntityState
	updated       map[string]time.Time
	entityAdapter entity.Adapter
}

// Old returns the number of objects that have updates to be extracted then flushed to disk/cache.
func (b *builder) Old(timeFilter time.Time) uint64 {
	if len(b.pending) == 0 {
		return uint64(0)
	}

	count := uint64(0)

	for _, updated := range b.updated {
		if updated.Before(timeFilter) {
			count++
		}
	}

	return count
}

// Extract all pending operations: new entities, updated entities.
// When calling Extract, dirty entities are marked as cleaned so the next
// call will not return them again.
func (b *builder) Extract(timeFilter time.Time) (map[string]EntityState, map[string]EntityState) {
	inserts := make(map[string]EntityState)
	updates := make(map[string]EntityState)

	for id, entityState := range b.pending {
		if entityState.State == PendingOK {
			continue
		}

		// We need to remove the local cache so we get fresh data from the bulk insert/update
		switch entityState.State {
		case PendingInsert:
			inserts[id] = entityState
			b.delLocal(id)
		case PendingUpdate:
			if entityState.Updated.Before(timeFilter) {
				updates[id] = entityState
				b.delLocal(id)
			}
		}
	}

	// Split inserts and updates since inserts must be done before updates.
	return inserts, updates
}

// reset internal states
func (b *builder) reset() {
	b.pending = make(map[string]EntityState)
	b.pendingOk = make(map[string]EntityState)
	b.updated = make(map[string]time.Time)
}

func (b *builder) getPending(id string) (EntityState, bool) {
	entityState, exists := b.pending[id]
	return entityState, exists
}

func (b *builder) getPendingOk(id string) (EntityState, bool) {
	entityState, exists := b.pendingOk[id]
	return entityState, exists
}

// get the entity if it exists
func (b *builder) get(id string) (EntityState, bool) {
	// local data that has not yet expired (extracted)
	entityState, existsLocal := b.getPending(id)
	if existsLocal {
		return entityState, true
	}

	entityState, existsLocal = b.getPendingOk(id)
	if existsLocal {
		return entityState, true
	}

	// remote data
	entity, exists := b.entityAdapter.Get(id)
	if exists {
		es := EntityState{
			State:   PendingOK,
			Entity:  entity,
			Updated: time.Now(),
		}
		b.setLocal(es)
		return es, true
	}
	return EntityState{}, false
}

// setLocal local cache with given entityState for further updates
func (b *builder) setLocal(entityState EntityState) {
	id := entityState.Entity.ID

	if entityState.State != PendingOK {
		b.pending[id] = entityState
		b.updated[id] = entityState.Updated
	} else {
		b.pendingOk[id] = entityState
		delete(b.pending, id)
		delete(b.updated, id)
	}
}

// delLocal entries of the given entity through it's ID
func (b *builder) delLocal(id string) {
	delete(b.pending, id)
	delete(b.updated, id)
	delete(b.pendingOk, id)
}

// Handle the given event to create or update entites linked to.
// You must call Enrich yourself to link extra infos.
func (b *builder) Handle(event types.Event) *types.Entity {
	ctxInfos := event.GenerateContextInformations()
	var eventEntity *types.Entity

	for _, ctxInfo := range ctxInfos {
		entityState, exists := b.get(ctxInfo.ID)
		if exists {
			impacts := ctxInfo.Impacts.List()
			depends := ctxInfo.Depends.List()

			entityState, ui := b.addImpacts(entityState, impacts...)
			entityState, ud := b.addDepends(entityState, depends...)

			if ui || ud {
				if entityState.State == PendingOK {
					entityState.State = PendingUpdate
				}
				entityState.Updated = time.Now()
				b.setLocal(entityState)
			}
		} else {
			newEntity := ctxInfo.NewEntity()
			entityState = b.Create(newEntity)
			b.setLocal(entityState)
		}

		if entityState.Entity.Type == event.SourceType {
			eventEntity = &entityState.Entity
		}
	}

	return eventEntity
}

// Create the entity without updating impacts/depends
func (b *builder) Create(entity types.Entity) EntityState {
	return EntityState{
		Entity:  entity,
		State:   PendingInsert,
		Updated: time.Now(),
	}
}

// addImpacts to the entity found with id.
// Returns true if an update has been queued.
func (b *builder) addImpacts(entityState EntityState, impacts ...string) (EntityState, bool) {
	updated := false
	for _, impact := range impacts {
		if entityState.Entity.Impacts.Exists(impact) {
			continue
		}

		entityState.Entity.Impacts.Add(impact)

		updated = true
	}
	return entityState, updated
}

// addDepends to the entity found with id.
// Returns true if an update has been queued
func (b *builder) addDepends(entityState EntityState, depends ...string) (EntityState, bool) {
	updated := false
	for _, depend := range depends {
		if entityState.Entity.Depends.Exists(depend) {
			continue
		}

		entityState.Entity.Depends.Add(depend)
		updated = true
	}
	return entityState, updated
}

// Enrich extra informations to entity
//
// param event is the event that carry the explicit keys
// param source is the raw event as byte array
// param ef is the EnrichFields which controls which fields are allowed to enrich extra infos
//
// Returns the modified entity
// Returns true if new infos has been added or modified
func (b *builder) Enrich(event types.Event, ef EnrichFields) *types.Entity {
	eid := event.GetEID()
	entityState, exists := b.get(eid)

	if !exists {
		return nil
	}

	updated := false

	for key, val := range event.ExtraInfos {
		if !ef.Allow(key) {
			continue
		}

		value, err := types.InterfaceToString(val)

		if err != nil {
			log.Printf("enrich context from event: key: %v, err: %v", key, err)
		}

		info := types.NewInfo(key, "", value)
		oldVal, exists := entityState.Entity.Infos[info.Name]

		if !exists || (exists && oldVal.Value != info.Value) {
			updated = true
			entityState.Entity.Infos[info.Name] = info
		}
	}

	if updated {
		if entityState.State == PendingOK {
			entityState.State = PendingUpdate
		}
		entityState.Updated = time.Now()
		b.setLocal(entityState)
	}

	return &entityState.Entity
}

// NewBuilder creates a new context builder.
func NewBuilder(entityAdapter entity.Adapter) Builder {
	b := builder{
		entityAdapter: entityAdapter,
	}
	b.reset()
	return &b
}
