package context

import (
	"log"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/entity"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
)

type center struct {
	entityAdapter entity.Adapter
	builder       Builder
	enableEnrich  bool
	updatedAfter  time.Duration
	maxOldObjects uint64
}

func (e *center) Handle(event types.Event, ef EnrichFields) *types.Entity {
	eventEntity := e.builder.Handle(event)
	if e.enableEnrich {
		eventEntity = e.builder.Enrich(event, ef)
	}

	timeFilter := time.Now().Add(-e.updatedAfter)
	olds := e.builder.Old(timeFilter)

	if olds >= e.maxOldObjects {
		//log.Printf("old objects: %d", olds)
		e.Flush(timeFilter)
	}

	return eventEntity
}

func (e *center) Flush(timeFilter time.Time) {
	inserts, updates := e.builder.Extract(timeFilter)

	if err := e.entityAdapter.FlushBulk(); err != nil {
		log.Printf("entity pre-flush: %v", err)
	}

	inserted := 0
	for _, entityState := range inserts {
		inserted++
		if err := e.entityAdapter.BulkInsert(entityState.Entity); err != nil {
			log.Printf("entity insertion: %v", err)
		}
	}

	updated := 0
	for _, entityState := range updates {
		updated++
		if err := e.entityAdapter.BulkUpdate(entityState.Entity); err != nil {
			log.Printf("entity update: %v", err)
		}
	}

	//log.Printf("inserted: %d", inserted)
	//log.Printf("updated: %d", updated)

	if err := e.entityAdapter.FlushBulk(); err != nil {
		log.Printf("entity post-flush: %v", err)
	}
}

// NewEnrichmentCenter creates an instance of the high level context creation api.
// param maxOldObjects: each Handle() call will check for objects that has not received updates since 10 seconds and
//    handle flush through entityAdapter
// param enableEnrich: enable extra infos enrichment
func NewEnrichmentCenter(maxOldObjects uint64, enableEnrich bool, entityAdapter entity.Adapter) EnrichmentCenter {
	builder := NewBuilder(entityAdapter)
	return &center{
		builder:       builder,
		entityAdapter: entityAdapter,
		enableEnrich:  enableEnrich,
		updatedAfter:  time.Second * 10,
		maxOldObjects: maxOldObjects,
	}
}
