package debug

import (
	"fmt"
	"log"
	"os"
	"path/filepath"
	"runtime"
	"runtime/pprof"
)

const (
	envEnable = "CPS_DEBUG_PPROF_ENABLE"
	envCPU    = "CPS_DEBUG_PPROF_CPU"
	envMemory = "CPS_DEBUG_PPROF_MEMORY"
)

var globalTrace *Trace

// Trace ...
type Trace struct {
	cpustarted bool
	memstarted bool
	memoutput  *os.File
}

// NewTrace creates a trace object
func NewTrace() Trace {
	return Trace{}
}

// StartGlobalTrace starts tracing in a global state.
// call defer StopGlobalTrace() and finish your program gracefuly.
func StartGlobalTrace() {
	if globalTrace == nil {
		t := NewTrace()
		globalTrace = &t
	}

	globalTrace.Start()
}

// StopGlobalTrace ends tracing
func StopGlobalTrace() {
	if globalTrace != nil {
		globalTrace.Stop()
	}
}

// Enable reads CPU_DEBUG_PPROF_ENABLE env var and returns true
// only if it equals "1"
func (t *Trace) Enable() bool {
	return os.Getenv(envEnable) == "1"
}

// CPUFile reads CPS_DEBUG_PPROF_CPU en var
func (t *Trace) CPUFile() (string, error) {
	fpath := os.Getenv(envCPU)

	if fpath == "" {
		return "", fmt.Errorf("empty %s env var", envCPU)
	}

	fpath, _ = filepath.Abs(fpath)

	return fpath, nil
}

// MemoryFile reads CPS_DEBUG_PPROF_MEMORY env var
func (t *Trace) MemoryFile() (string, error) {
	fpath := os.Getenv(envMemory)

	if fpath == "" {
		return "", fmt.Errorf("empty %s env var", envMemory)
	}

	fpath, _ = filepath.Abs(fpath)

	return fpath, nil
}

// StartCPU ...
func (t *Trace) StartCPU() error {
	fpath, err := t.CPUFile()
	if err != nil {
		return fmt.Errorf("pprof cpu trace conf: %v", err)
	}

	fh, err := os.Create(fpath)
	if err != nil {
		return fmt.Errorf("pprof cpu trace create: %v", err)
	}

	err = pprof.StartCPUProfile(fh)
	if err != nil {
		return fmt.Errorf("pprof cpu trace start: %v", err)
	}

	log.Printf("CPU profiling enabled on file: %s", fpath)

	return nil
}

// StartMemory ...
func (t *Trace) StartMemory() error {
	fpath, err := t.MemoryFile()
	if err != nil {
		return fmt.Errorf("pprof mem trace conf: %v", err)
	}

	fh, err := os.Create(fpath)
	if err != nil {
		return fmt.Errorf("pprof mem trace create: %v", err)
	}

	t.memoutput = fh

	log.Printf("Memory profiling enabled on file: %s", fpath)

	return nil
}

// Start will start cpu and memory profiling if required variables are set:
//
// CPS_DEBUG_PPROF_ENABLE: equals to 1 starts profiling
// CPS_DEBUG_PPROF_CPU: path to profile output
// CPS_DEBUG_PPROF_MEMORY: path to profile output
func (t *Trace) Start() error {
	if !t.Enable() {
		return nil
	}

	log.Println("Profiling ENABLED")

	cpuerr := t.StartCPU()
	memerr := t.StartMemory()

	if cpuerr != nil && memerr != nil {
		log.Printf("cpu error: %v", cpuerr)
		log.Fatalf("mem error: %v", memerr)
	}

	if cpuerr != nil {
		log.Printf("cpu error: %v", cpuerr)
	} else {
		t.cpustarted = true
	}

	if memerr != nil {
		log.Printf("mem error: %v", memerr)
	} else {
		t.memstarted = true
	}

	return nil
}

// Stop started tracings
func (t *Trace) Stop() {
	if t.cpustarted {
		pprof.StopCPUProfile()
		t.cpustarted = false
	}

	if t.memstarted {
		runtime.GC()
		if err := pprof.WriteHeapProfile(t.memoutput); err != nil {
			log.Printf("mem error: cannot write heap profile: %v", err)
		}
		t.memoutput.Close()
		t.memoutput = nil
		t.memstarted = false
	}
}
