package canopsis

import (
	"errors"
	"log"
	"os"
	"os/signal"
	"runtime/debug"
	"syscall"
	"time"

	"github.com/streadway/amqp"
)

// Returned values on exit
const (
	ExitOK = iota
	ExitPanic
	ExitEngine
)

// Engine makes engines more predictible in the way they are built.
// Used with StartEngine(), goroutine spawn is handled for you
// and you only have to implement your working logic.
//
// You can use the DefaultEngine struct to avoid implementing
// Initialize(), Stop(), Continue(), Confirm*Stop() and Has*Process()
type Engine interface {
	// PeriodicalWaitTime returns the duration to wait between
	// two run of PeriodicalProcess()
	PeriodicalWaitTime() time.Duration

	// Continue returns the state of the engine.
	// true to continue running
	// false to stop as soon as possible, meaning when functions finish
	// their job.
	Continue() bool

	// ConsumerChan must return a new channel from amqp.Channel.Consume()
	// The worker process will stop looping when the channel is closed,
	// so you must close yourself all channels.
	ConsumerChan() (<-chan amqp.Delivery, error)

	// Initialize anything you want. If err != nil then the Engine will not start.
	Initialize() error
	Started()

	// WorkerProcess must implement the actual consumer processing.
	WorkerProcess(amqp.Delivery)

	// PeriodicalProcess must implement the actual "beat processing".
	// Prefer using channels that you will close for long
	// PeriodicalWaitTime() values.
	PeriodicalProcess()

	SetWaitStopChan(chan os.Signal)
	AskStop(int)
	// Stop handles stopping the engine, and also waiting
	// for goroutines to finish.
	Stop()

	// ConfirmPeriodicalStop is useful to mutate a variable
	// or an internal chan to confirm that the Periodical process
	// is now stopped.
	ConfirmPeriodicalStop()
	ConfirmWorkerStop()

	RunPeriodicalProcess() bool
	RunWorkerProcess() bool

	// Do NOT override when using DefaultEngine.
	// Use this method to send a ConfirmStop without doing anything else.
	AcknowledgeConfirmStop()

	// RecoverWorker and RecoverPeriodical handle panic recovery for
	// WorkerProcess and PeriodicalProcess respectively.
	// The default implementation prints the stack and ask engine to stop.
	RecoverWorker()
	RecoverPeriodical()
}

// DefaultEngine provides basic functions and will behave as an Engine
// with PeriodicalProcess and WorkerProcess.
//
// Stop() will automatically close the Sub channel and waiting
// for confirmations to come on ConfirmChan.
type DefaultEngine struct {
	cont          bool
	sigchan       chan os.Signal
	stopAsked     bool
	returnValue   int
	ConfirmChan   chan bool
	Sub           *amqp.Channel
	Sleep         time.Duration
	RunWorker     bool
	RunPeriodical bool
	Debug         bool
}

// NewDefaultEngine returns an default engine implementation.
// Check engine_test.go in this package to see how you can use it.
func NewDefaultEngine(sleep time.Duration, runworker, runperiodical bool, sub *amqp.Channel) DefaultEngine {
	return DefaultEngine{
		cont:          true,
		returnValue:   ExitOK,
		Sleep:         sleep,
		ConfirmChan:   make(chan bool, 2),
		Sub:           sub,
		RunWorker:     runworker,
		RunPeriodical: runperiodical,
	}
}

// Initialize does nothing here besides returning nil.
func (de *DefaultEngine) Initialize() error {
	return nil
}

// Started is called before StartEngine waits for stop signal.
func (de *DefaultEngine) Started() {
	log.Printf("Engine started")
}

// Continue returns the DefaultEngine.cont flag.
func (de *DefaultEngine) Continue() bool {
	return de.cont
}

// PeriodicalWaitTime returns the Sleep attribute.
func (de *DefaultEngine) PeriodicalWaitTime() time.Duration {
	return de.Sleep
}

// RunPeriodicalProcess returns DefaultEngine.RunPeriodical
func (de DefaultEngine) RunPeriodicalProcess() bool {
	return de.RunPeriodical
}

// RunWorkerProcess returns DefaultEngine.RunWorker
func (de DefaultEngine) RunWorkerProcess() bool {
	return de.RunWorker
}

// SetWaitStopChan keeps in memory our sigint/term chan.
func (de *DefaultEngine) SetWaitStopChan(sigchan chan os.Signal) {
	de.sigchan = sigchan
}

// AskStop sends SIGINT into the channel where StartEngine waits for sigint/term,
// leading to a proper engine stop.
func (de *DefaultEngine) AskStop(state int) {
	if !de.stopAsked {
		de.returnValue = state
		de.sigchan <- syscall.SIGINT
	}
	de.stopAsked = true
}

// Stop set cont flag to false, close the Sub channel then wait
// for two values from ConfirmChan.
func (de *DefaultEngine) Stop() {
	log.Print("engine stop called, waiting for processes to finish.")
	de.cont = false
	if de.Sub != nil {
		de.Sub.Close()
	}
	<-de.ConfirmChan
	<-de.ConfirmChan

	os.Exit(de.returnValue)
}

// AcknowledgeConfirmStop sends true to ConfirmChan once for each call.
// Do NOT override this method when using DefaultEngine in your struct.
func (de *DefaultEngine) AcknowledgeConfirmStop() {
	de.ConfirmChan <- true
}

// ConfirmWorkerStop sends one value to ConfirmChan.
func (de *DefaultEngine) ConfirmWorkerStop() {
	de.AcknowledgeConfirmStop()
}

// ConfirmPeriodicalStop sends one value to ConfirmChan.
func (de *DefaultEngine) ConfirmPeriodicalStop() {
	de.AcknowledgeConfirmStop()
}

// DebugPrint print a log only if the engine is in debug mode
func (de *DefaultEngine) DebugPrint(fmt string, params ...interface{}) {
	if de.Debug {
		if len(params) > 0 {
			log.Printf("DEBUG | "+fmt, params...)
		} else {
			log.Print("DEBUG | " + fmt)
		}
	}
}

// StartEngine handles starting the WorkerProcess and PeriodicalProcess
// of your Engine.
// It first calls Engine.Initialize(), returns an error if any, then
// proceed to signals binding.
//
// Engine.PeriodicalProcess and Engine.WorkerProcess are launched into
// separate goroutines.
//
// Only SIGTERM or SIGINT will trigger the Engine.Stop() method.
//
// waitChan is optional: if nil a chan will be instanciated and
// managed for you. This is mainly for tests.
func StartEngine(engine Engine, waitChan *chan os.Signal) error {
	if !engine.RunPeriodicalProcess() && !engine.RunWorkerProcess() {
		return errors.New("impossible engine configuration: must have at least one Process()")
	}

	err := engine.Initialize()
	if err != nil {
		return err
	}

	if waitChan == nil {
		w := make(chan os.Signal, 1)
		waitChan = &w
	}

	engine.SetWaitStopChan(*waitChan)
	signal.Notify(*waitChan, syscall.SIGINT, syscall.SIGTERM, syscall.SIGKILL)

	if engine.RunWorkerProcess() {
		channel, err := engine.ConsumerChan()
		if err != nil {
			return err
		}
		go routineWorker(engine, channel)
	} else {
		engine.AcknowledgeConfirmStop()
	}

	if engine.RunPeriodicalProcess() {
		go routinePeriodical(engine)
	} else {
		engine.AcknowledgeConfirmStop()
	}

	engine.Started()

	<-*waitChan
	engine.Stop()
	return nil
}

// RecoverWorker handle panic recovery in the worker goroutine
func (de *DefaultEngine) RecoverWorker() {
	if r := recover(); r != nil {
		log.Printf("worker recovered from panic: %v", r)
		debug.PrintStack()
		de.AskStop(ExitPanic)
		de.ConfirmWorkerStop()
	}
}

// RecoverPeriodical handle panic recovery in the periodical goroutine
func (de *DefaultEngine) RecoverPeriodical() {
	if r := recover(); r != nil {
		log.Printf("periodical recovered from panic: %v", r)
		debug.PrintStack()
		de.AskStop(ExitPanic)
		de.ConfirmPeriodicalStop()
	}
}

func routineWorker(engine Engine, channel <-chan amqp.Delivery) {
	defer engine.RecoverWorker()
	timeout := make(chan bool, 1)
	go func() {
		for engine.Continue() {
			time.Sleep(time.Second * 1)
			timeout <- true
		}
	}()

	for engine.Continue() {
		select {
		case msg := <-channel:
			if len(msg.Body) > 0 {
				engine.WorkerProcess(msg)
			}
		case <-timeout:
		}
	}

	engine.AskStop(ExitOK)
	engine.ConfirmWorkerStop()
}

// watch engine.Continue every 2 second, during a maximum amount of time
// given by engine.PeriodicalWaitTime(), and return engine.Continue().
func waitAndCheck(engine Engine) bool {
	sleepCountDown := engine.PeriodicalWaitTime()
	for sleepCountDown > 0 && engine.Continue() {
		wait := time.Second * 2
		time.Sleep(wait)
		sleepCountDown = sleepCountDown - wait
	}
	return engine.Continue()
}

func routinePeriodical(engine Engine) {
	defer engine.RecoverPeriodical()
	for waitAndCheck(engine) {
		//for engine.Continue() {
		engine.PeriodicalProcess()
	}

	engine.AskStop(ExitOK)
	engine.ConfirmPeriodicalStop()
}
