package entity

import (
	"fmt"

	"git.canopsis.net/canopsis/go-revolution/lib/errt"

	"git.canopsis.net/canopsis/go-revolution/lib/mongo"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/cache"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo/bulk"
	"github.com/go-redis/redis"
	mgo "github.com/globalsign/mgo"
	"github.com/globalsign/mgo/bson"
)

// Entities constants
const (
	EntityCollectionName = "default_entities"
)

// DBAdapter ...
type DBAdapter struct {
	Collection *mgo.Collection
	Cache      cache.Cache
	mbulk      bulk.Bulk
}

// NewAdapter gives the correct entity adapter. Give nil to the redis client
// and it will create a new redis.Client with the dedicated redis database for entities.
func NewAdapter(session *mgo.Session, redis *redis.Client) DBAdapter {
	entityCache := NewCache(redis)
	cache := cache.NewThreadSafeCache(&entityCache)
	collection := session.DB(cps.DbName).C(EntityCollectionName)
	mbulk := bulk.NewSafe(bulk.New(collection, bulk.BulkSizeMax))
	return DBAdapter{
		Collection: collection,
		Cache:      cache,
		mbulk:      mbulk,
	}
}

// GetEntityByID finds an Entity from is eid
func (e *DBAdapter) GetEntityByID(id string) (types.Entity, error) {
	entity, exists := e.Cache.Get(id)
	if exists {
		return entity.(types.Entity), nil
	}

	var ent types.Entity
	err := mongo.WrapError(e.Collection.FindId(id).One(&ent))
	_, notfound := err.(errt.NotFound)
	if err != nil && !notfound {
		return ent, err
	}

	if notfound {
		e.Cache.Drop(id)
		return ent, err
	}

	ent.EnsureInitialized()

	e.Cache.Set(ent)
	return ent, nil
}

// Get is the same as GetEntityByID
func (e *DBAdapter) Get(id string) (types.Entity, bool) {
	entity, err := e.GetEntityByID(id)
	_, notfound := err.(errt.NotFound)
	entity.EnsureInitialized()
	if notfound {
		return entity, false
	} else if err != nil {
		return entity, false
	}
	return entity, true
}

// Insert add a new entity.
func (e *DBAdapter) Insert(entity types.Entity) error {
	e.Cache.Set(entity)
	return mongo.WrapError(e.Collection.Insert(entity))
}

// Update updates an existing entity or creates a new one in db.
func (e *DBAdapter) Update(entity types.Entity) error {
	e.Cache.Set(entity)
	return mongo.WrapError(e.Collection.UpdateId(entity.ID, entity))
}

// Remove an existing entity in db.
func (e *DBAdapter) Remove(entity types.Entity) error {
	e.Cache.Drop(entity.ID)
	return mongo.WrapError(e.Collection.Remove(bson.M{"_id": entity.ID}))
}

// BulkInsert insert entities in bulk.
// Thread safe.
func (e *DBAdapter) BulkInsert(entity types.Entity) error {
	e.Cache.Set(entity)
	_, err := e.mbulk.AddInsert(
		bulk.NewInsert(entity),
	)
	return err
}

// BulkUpdate update entities in bulk
// Thread safe.
func (e *DBAdapter) BulkUpdate(entity types.Entity) error {
	e.Cache.Set(entity)
	_, err := e.mbulk.AddUniqUpdate(
		entity.ID, bulk.NewUpdate(bson.M{"_id": entity.ID}, entity),
	)
	return err
}

// FlushBulk force all bulk caches to be written.
func (e *DBAdapter) FlushBulk() error {
	errInsert := e.FlushBulkInsert()
	errUpdate := e.FlushBulkUpdate()

	if errInsert != nil {
		return fmt.Errorf("entity adapter flushbulk insert: %v", errInsert)
	}
	if errUpdate != nil {
		return fmt.Errorf("entity adapter flushbulk update: %v", errUpdate)
	}
	return nil
}

// FlushBulkInsert force insert cache to be sent to Mongo.
// This function is thread safe.
// TODO: fusion with FlushBulkUpdate() in one single algorythm ?
func (e *DBAdapter) FlushBulkInsert() error {
	_, err := e.mbulk.PerformInserts()
	if err != nil {
		return fmt.Errorf("entities bulk flush inserts: %v", err)
	}
	return nil
}

// FlushBulkUpdate force update cache to be sent to Mongo.
// This function is thread safe.
// TODO: fusion with FlushBulkInsert() ?
func (e *DBAdapter) FlushBulkUpdate() error {
	_, err := e.mbulk.PerformUniqUpdates()
	if err != nil {
		return fmt.Errorf("entities bulk flush updates: %v", err)
	}
	return nil
}
