package entity_test

import (
	"fmt"
	"strconv"
	"testing"

	"git.canopsis.net/canopsis/go-revolution/lib/errt"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/entity"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo"
	"git.canopsis.net/canopsis/go-revolution/lib/testutils"
	. "github.com/smartystreets/goconvey/convey"
)

func TestEntityDBAdapter(t *testing.T) {
	Convey("Given a mongo session", t, func() {
		s, err := mongo.NewSession(mongo.Timeout * 10)
		So(err, ShouldBeNil)

		collection := s.DB(cps.DbName).C(entity.EntityCollectionName)
		err = testutils.MongoDrop(collection)
		So(err, ShouldBeNil)

		Convey("I can get an adapter", func() {
			ad := entity.NewAdapter(s, nil)
			ad.Cache.Flush()

			Convey("And check for an nonexistent entity", func() {
				_, err := ad.GetEntityByID("this is not this entity you are searching for")
				_, ok := err.(errt.NotFound)
				So(ok, ShouldBeTrue)
			})

			Convey("Set an entity in database", func() {
				ent := types.Entity{
					Enabled:       true,
					ID:            "squeeky/squeeke",
					Type:          "watcher",
					Depends:       types.NewDedupList(),
					EnableHistory: make([]types.CpsTime, 0),
					Impacts:       types.NewDedupList(),
				}

				err := s.DB(cps.DbName).C(entity.EntityCollectionName).Insert(ent)
				So(err, ShouldBeNil)

				Convey("Get it back", func() {
					dbentity, err := ad.GetEntityByID("squeeky/squeeke")

					So(err, ShouldBeNil)
					So(dbentity, ShouldNotBeNil)
					So(dbentity.ID, ShouldEqual, "squeeky/squeeke")
					So(dbentity.Type, ShouldEqual, "watcher")
					So(dbentity.Enabled, ShouldEqual, true)
				})
			})
		})
	})
}

// Verify that concurrency does not affect bulk operations
func TestBulkConcurrency(t *testing.T) {
	t.SkipNow()
	testutils.SkipLongIfSet(t)
	Convey("Given some bulk inserts in two goroutines", t, func() {
		s, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		collection := s.DB(cps.DbName).C(entity.EntityCollectionName)
		err = testutils.MongoDrop(collection)
		So(err, ShouldBeNil)
		ed := entity.NewAdapter(s, nil)
		err = ed.Cache.Flush()
		So(err, ShouldBeNil)
		err = ed.FlushBulk()
		So(err, ShouldBeNil)

		n, err := collection.Count()
		So(err, ShouldBeNil)
		So(n, ShouldEqual, 0)

		id := ""
		name := "Spamalot"

		finished := make(chan bool, 2)
		start := 0
		mid := 40000
		high := 80000

		writeEven := func() {
			entity := types.NewEntity(
				id,
				name,
				types.EntityTypeComponent,
				make(map[string]types.Info, 0),
				types.NewDedupList(),
				types.NewDedupList(),
			)
			for index := start; index < mid; index++ {
				entity.ID = strconv.Itoa(index)
				entity.Name = name + "_LOW_" + strconv.Itoa(index)
				ed.BulkInsert(entity)
			}
			finished <- true
		}
		writeOdd := func() {
			entity := types.NewEntity(
				id,
				name,
				types.EntityTypeComponent,
				make(map[string]types.Info, 0),
				types.NewDedupList(),
				types.NewDedupList(),
			)
			for index := mid; index < high; index++ {
				entity.ID = strconv.Itoa(index)
				entity.Name = name + "_HIGH_" + strconv.Itoa(index)
				ed.BulkInsert(entity)
			}
			finished <- true
		}

		test := func() {
			go writeOdd()
			go writeEven()
			<-finished
			<-finished
			err := ed.FlushBulk()
			if err != nil {
				panic(fmt.Errorf("test flushbulk on entity db adapter: %v", err))
			}
		}

		Convey("Then concurrent access on bulk insert should work", func() {
			So(test, ShouldNotPanic)
			count, err := ed.Collection.Count()
			So(err, ShouldBeNil)
			So(count, ShouldEqual, high)
		})
	})
}
