package entity

import (
	"fmt"
	"time"

	gogob "encoding/gob"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/cache"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/encoding"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/encoding/gob"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	cpsredis "git.canopsis.net/canopsis/go-revolution/lib/redis"
	"git.canopsis.net/canopsis/go-revolution/lib/utils"

	"github.com/go-redis/redis"
)

// NewBackend handle cache connection
func NewBackend() (*redis.Client, error) {
	return cpsredis.NewSession(cpsredis.CacheEntity)
}

// Cache ...
type Cache struct {
	backend *redis.Client
	encoder encoding.Encoder
	decoder encoding.Decoder
}

// NewCache creates the local cache and remote cache access for entitys.
func NewCache(redis *redis.Client) Cache {
	if redis == nil {
		nredis, err := NewBackend()
		utils.FailOnError(err, "entity cache redis")
		redis = nredis
	}
	gogob.Register(types.Entity{})
	c := Cache{
		backend: redis,
		encoder: gob.NewEncoder(),
		decoder: gob.NewDecoder(),
	}

	return c
}

// Get entity from cache if it exists. Returns the entity, true or entity, false
func (c *Cache) Get(id string) (interface{}, bool) {
	r := c.backend.Get(id)

	if r.Err() == redis.Nil {
		return nil, false
	}

	b, _ := r.Bytes()

	var entity types.Entity

	err := c.decoder.Decode(b, &entity)
	if err != nil {
		switch err.(type) {
		case encoding.DecodingError:
			c.Drop(id)
			return nil, false
		default:
			return nil, false
		}
	}

	entity.EnsureInitialized()

	return entity, true
}

// Flush entity cache
func (c *Cache) Flush() error {
	return c.backend.FlushDB().Err()
}

// Drop entities in local and remote cache. Use Entity.ID for ids.
func (c *Cache) Drop(ids ...string) error {
	return c.backend.Del(ids...).Err()
}

// Set the local and remote cache with this entity. If entity is nil, only
// the local cache is set with this value.
func (c *Cache) Set(entity cache.Cacheable) error {
	if entity != nil {
		b, err := c.encoder.Encode(entity)
		if err != nil {
			switch err.(type) {
			case encoding.DecodingError:
				return encoding.NewDecodingError(fmt.Errorf("entity cache set: %v", err))
			default:
				return fmt.Errorf("entity cache unexpected: %v", err)
			}
		}
		return c.backend.Set(entity.CacheID(), b, time.Hour*4).Err()
	}
	return nil
}
