package event_test

import (
	"fmt"
	"testing"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/testutils"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/alarm"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/entity"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/event"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo"
	. "github.com/smartystreets/goconvey/convey"
)

func TestNewEventDBAdapter(t *testing.T) {
	Convey("Setup", t, func() {
		db, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		defer db.Close()
		adba := alarm.NewService(db, nil, false)
		edba := entity.NewAdapter(db, nil)
		eventdba := event.NewEventDBAdapter(&adba, &edba)
		So(eventdba.AlarmService.Adapter.Collection.Name, ShouldEqual, alarm.AlarmCollectionName)
	})
}

func getTestEvent() types.Event {
	now := types.CpsTime{Time: time.Now()}
	return types.Event{
		Connector:     "godspeed_you",
		ConnectorName: "black_emperor",
		Component:     "yanqui",
		Resource:      "redeemer",
		EventType:     types.EventTypeCheck,
		SourceType:    types.SourceTypeResource,
		State:         types.AlarmStateOK,
		Timestamp:     now,
	}
}

func getTestAlarm() types.Alarm {
	stateStep := types.AlarmStep{
		Type:      "",
		Timestamp: types.CpsTime{Time: time.Now()},
		Author:    "Sleep",
		Message:   "The playground of the world",
		Value:     types.AlarmStateMajor,
	}

	v := types.AlarmValue{
		State:          &stateStep,
		Status:         nil,
		ACK:            nil,
		Canceled:       nil,
		Ticket:         nil,
		Resolved:       nil,
		Steps:          make(types.AlarmSteps, 0),
		Tags:           make([]string, 0),
		CreationDate:   types.CpsTime{Time: time.Now()},
		InitialOutput:  "",
		Snooze:         nil,
		HardLimit:      nil,
		Extra:          make(map[string]string),
		Connector:      "godspeed_you",
		ConnectorName:  "black_emperor",
		Component:      "yanqui",
		Resource:       "redeemer",
		LastUpdateDate: types.CpsTime{Time: time.Now()},
	}

	a := types.Alarm{
		ID:       "gybe",
		Time:     types.CpsTime{Time: time.Now()},
		EntityID: "redeemer/yanqui",
		Value:    v,
	}

	return a
}

func getEntity() types.Entity {
	return types.Entity{
		ID:            "redeemer/yanqui",
		Name:          "UXO",
		Impacts:       types.NewDedupList(),
		Depends:       types.NewDedupList(),
		EnableHistory: make([]types.CpsTime, 0),
		Measurements:  nil,
		Enabled:       true,
		Infos:         make(map[string]types.Info, 0),
		Type:          "moya",
	}
}

func cleanUpAlarm(alarm types.Alarm, as *alarm.Service) {
	as.Adapter.RemoveId(alarm.ID)
}

func cleanUpEntity(entity types.Entity, edba *entity.DBAdapter) {
	edba.Remove(entity)
}

func TestEnrich(t *testing.T) {
	Convey("Setup", t, func() {
		db, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)
		defer db.Close()
		defer testutils.MongoDrop(db.DB(cps.DbName).C(entity.EntityCollectionName))
		adba := alarm.NewService(db, nil, false)
		edba := entity.NewAdapter(db, nil)
		eventdba := event.NewEventDBAdapter(&adba, &edba)

		event := getTestEvent()
		event_id := fmt.Sprintf("%s/%s", event.Resource, event.Component)

		alarm := getTestAlarm()
		entity := getEntity()
		cleanUpAlarm(alarm, &adba)
		cleanUpEntity(entity, &edba)

		Convey("Then, with no alarm, there is not enrichment", func() {
			event, report := eventdba.Enrich(event)

			So(report.Error, ShouldBeNil)
			So(report.HasEntity, ShouldBeFalse)
			So(event.Alarm, ShouldBeNil)
			So(event.Entity, ShouldBeNil)
		})

		Convey("Then, with an entity, event is enriched with all informations", func() {
			err := edba.Insert(entity)
			So(err, ShouldBeNil)

			event, report := eventdba.Enrich(event)
			So(report.Error, ShouldBeNil)
			So(event.Entity, ShouldNotBeNil)

			db_entity, err := edba.GetEntityByID(event_id)
			So(err, ShouldBeNil)
			So(db_entity.Name, ShouldEqual, event.Entity.Name)
			So(db_entity.ID, ShouldEqual, event.Entity.ID)
		})
	})
}
