package canopsis

import (
	"fmt"

	cpsamqp "git.canopsis.net/canopsis/go-revolution/lib/amqp"
	cpsinflux "git.canopsis.net/canopsis/go-revolution/lib/influx"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo"
	cpsredis "git.canopsis.net/canopsis/go-revolution/lib/redis"
	redismod "github.com/go-redis/redis"
	influxmod "github.com/influxdata/influxdb/client/v2"
	amqpmod "github.com/streadway/amqp"
	"github.com/globalsign/mgo"
)

// CpsSessionsConf ease the use of NewCpsSessions.
// Do not set field to false if you don't need a session
// since bool default value is false.
type CpsSessionsConf struct {
	Mongo   bool
	Redis   bool
	Amqp    bool
	AmqpPub bool
	AmqpSub bool
	RedisDB int
	Influx  bool
}

// CpsSessions regroups all sessions. Do not add anything else than pointers.
type CpsSessions struct {
	Mongo   *mgo.Session
	Redis   *redismod.Client
	Amqp    *amqpmod.Connection
	AmqpPub *amqpmod.Channel
	AmqpSub *amqpmod.Channel
	Influx  influxmod.Client
}

// NewCpsSessions handles the tedious work of creating every required sessions
// from a CpsSessionsConf struct.
//
// Note: if you require AmqpPub/Sub but not Amqp, a new Amqp session *is* created.
//
// Note2: if you add some sessions, do not forget to modify Close() too!
func NewCpsSessions(cfg CpsSessionsConf) (*CpsSessions, error) {
	// All pointers will be nil, just set those we want.
	sessions := CpsSessions{}

	if cfg.Mongo {
		mongo, err := mongo.NewSession(mongo.Timeout)
		if err != nil {
			return nil, fmt.Errorf("MONGO: %v", err)
		}
		sessions.Mongo = mongo
	}

	if cfg.Redis {
		r, err := cpsredis.NewSession(cfg.RedisDB)
		if err != nil {
			return nil, fmt.Errorf("REDIS: %v", err)
		}
		sessions.Redis = r
	}

	if cfg.Amqp || cfg.AmqpPub || cfg.AmqpSub {
		amqp, err := cpsamqp.NewSession()
		if err != nil {
			return nil, fmt.Errorf("AMQP: %v", err)
		}
		sessions.Amqp = amqp

		if cfg.AmqpPub {
			amqppub, err := amqp.Channel()
			if err != nil {
				return nil, fmt.Errorf("AMQP PUB: %v", err)
			}
			sessions.AmqpPub = amqppub
		}

		if cfg.AmqpSub {
			amqpsub, err := amqp.Channel()
			if err != nil {
				return nil, fmt.Errorf("AMQP SUB: %v", err)
			}
			sessions.AmqpSub = amqpsub
		}
	}

	if cfg.Influx {
		influx, err := cpsinflux.NewSession()
		if err != nil {
			return nil, fmt.Errorf("INFLUX: %v", err)
		}
		sessions.Influx = influx
	}

	return &sessions, nil
}

// Close handles closing every single session, if created.
// It does not rely on CpsSessionsConf.
func (cs *CpsSessions) Close() {
	if cs.Mongo != nil {
		cs.Mongo.Close()
		cs.Mongo = nil
	}
	if cs.Redis != nil {
		cs.Redis.Close()
		cs.Redis = nil
	}
	if cs.Amqp != nil {
		if cs.AmqpPub != nil {
			cs.AmqpPub.Close()
			cs.AmqpPub = nil
		}

		if cs.AmqpSub != nil {
			cs.AmqpSub.Close()
			cs.AmqpSub = nil
		}

		cs.Amqp.Close()
		cs.Amqp = nil
	}
	if cs.Influx != nil {
		cs.Influx = nil
	}
}
