package types

import (
	"fmt"
	"time"
)

// Alarm consts
const (
	AlarmStepCropMinStates   = 20
	AlarmStepsHardLimit      = 2000
	AlarmStepTypeSateCounter = "statecounter"
)

// AlarmStep represents a generic step used in an alarm.
type AlarmStep struct {
	Type         string    `bson:"_t" json:"_t"`
	Timestamp    CpsTime   `bson:"t" json:"t"`
	Author       string    `bson:"a" json:"a"`
	Message      string    `bson:"m" json:"m"`
	Value        CpsNumber `bson:"val" json:"val"`
	StateCounter *PyMap    `bson:"statecounter,omitempty" json:"statecounter,omitempty"`
}

// AlarmSteps is a sortable implementation of []*AlarmStep. Used for sorting
// steps in some functions. Implements sort.Interface
type AlarmSteps []AlarmStep

// Add handle adding a step to the list
func (s *AlarmSteps) Add(step AlarmStep) error {
	if len(*s) >= AlarmStepsHardLimit {
		return fmt.Errorf("max number of steps reached: %v", AlarmStepsHardLimit)
	}

	*s = append(*s, step)
	return nil
}

// Crop steps by replacing all stateinc and statedec steps with statecounter-s
// Returns the newly added statecounter step.
// param fromFirstState: crop from the first step with the given AlarmState
// param cropNum: crop only if we have at least cropNum steps with type AlarmStepStateIncrease or AlarmStepStateDecrease
// param author: author for the new step
// param message: message for the new step
func (s *AlarmSteps) Crop(fromFirstState CpsNumber, cropNum int, author, message string) (AlarmStep, bool) {
	update := false
	changeSteps := make(AlarmSteps, 0)
	cleanedSteps := make(AlarmSteps, 0)
	fstateIdx := -1

	// prepare crop
	for i, step := range *s {
		if step.Type != AlarmStepStateDecrease && step.Type != AlarmStepStateIncrease {
			cleanedSteps.Add(step)
		} else if step.Value == fromFirstState && fstateIdx == -1 {
			fstateIdx = i
		}

		if fstateIdx > -1 {
			if step.Type == AlarmStepStateIncrease || step.Type == AlarmStepStateDecrease {
				changeSteps.Add(step)
			}
		}
	}

	if fstateIdx < 0 || len(changeSteps) < cropNum {
		return AlarmStep{}, false
	}

	cropCounters := NewPyMap()
	for i := 0; i < len(changeSteps); i++ {
		t := changeSteps[i].Type
		cropCounters[t] = cropCounters.Get(t, 0).(int) + 1

		s := fmt.Sprintf("state:%d", changeSteps[i].Value)
		cropCounters[s] = cropCounters.Get(s, 0).(int) + 1
	}

	// crop steps
	newStep := AlarmStep{
		Author:    author,
		Message:   message,
		Timestamp: CpsTime{Time: time.Now()},
		Type:      AlarmStepTypeSateCounter,
	}
	npm := NewPyMap()
	newStep.StateCounter = &npm

	for category, count := range cropCounters {
		nc := newStep.StateCounter.Get(category, 0).(int) + count.(int)
		newStep.StateCounter.Set(category, nc)
	}

	cleanedSteps.Add(newStep)

	*s = cleanedSteps

	update = true
	return newStep, update
}

// Last returns the last step, if any, or returns an error.
func (s AlarmSteps) Last() (AlarmStep, error) {
	if len(s) > 0 {
		return s[len(s)-1], nil
	}
	return AlarmStep{}, fmt.Errorf("no step")
}

// AlarmValue represents a full description of an alarm.
type AlarmValue struct {
	ACK            *AlarmStep        `bson:"ack,omitempty"`
	Canceled       *AlarmStep        `bson:"canceled,omitempty"`
	Snooze         *AlarmStep        `bson:"snooze,omitempty"`
	State          *AlarmStep        `bson:"state,omitempty"`
	Status         *AlarmStep        `bson:"status,omitempty"`
	Ticket         *AlarmTicket      `bson:"ticket,omitempty"`
	Steps          AlarmSteps        `bson:"steps"`
	Component      string            `bson:"component"`
	Connector      string            `bson:"connector"`
	ConnectorName  string            `bson:"connector_name"`
	CreationDate   CpsTime           `bson:"creation_date"`
	DisplayName    string            `bson:"display_name"`
	Extra          map[string]string `bson:"extra"`
	HardLimit      *CpsNumber        `bson:"hard_limit,omitempty"`
	InitialOutput  string            `bson:"initial_output"`
	LastUpdateDate CpsTime           `bson:"last_update_date"`
	LastEventDate  CpsTime           `bson:"last_event_date"`
	Resource       string            `bson:"resource,omitempty"`
	Resolved       *CpsTime          `bson:"resolved,omitempty"`
	Tags           []string          `bson:"tags"`
}

// AlarmTicket step is distinct from generic alarm step because value is a string
// TODO: move string value to message (and in py and js too)
type AlarmTicket struct {
	Type      string  `bson:"_t" json:"_t"`
	Timestamp CpsTime `bson:"t" json:"t"`
	Author    string  `bson:"a" json:"a"`
	Message   string  `bson:"m" json:"m"`
	Value     string  `bson:"val" json:"val"`
}

// NewTicket creates a Ticket Step from a normal step
// TODO: annihilate this heresy
func (s AlarmStep) NewTicket(value string) AlarmTicket {
	return AlarmTicket{
		Author:    s.Author,
		Message:   value,
		Timestamp: s.Timestamp,
		Type:      s.Type,
		Value:     value,
	}
}
