package types_test

import (
	"testing"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/alarm"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	"git.canopsis.net/canopsis/go-revolution/lib/mongo"
	. "github.com/smartystreets/goconvey/convey"
)

func TestAlarmStepsCrop(t *testing.T) {
	Convey("Setup", t, func() {
		steps := make(types.AlarmSteps, 0)
		s1 := types.AlarmStep{
			Type:  types.AlarmStepStateIncrease,
			Value: 2.0,
		}
		s2 := types.AlarmStep{
			Type:  types.AlarmStepStateDecrease,
			Value: 1.0,
		}
		si := types.AlarmStep{
			Type:  types.AlarmStepAck,
			Value: 0.0,
		}
		for i := 0; i < 10; i++ {
			steps.Add(s1)
			steps.Add(s2)
		}

		Convey("Crop does not work for the first 21 steps", func() {
			steps.Add(si)
			_, update := steps.Crop(s1.Value, 40, "test", "test")
			So(update, ShouldBeFalse)
			Convey("Crop does work for the 41 steps", func() {
				for i := 0; i < 10; i++ {
					steps.Add(s1)
					steps.Add(s2)
				}

				nstep, update := steps.Crop(s1.Value, 40, "test", "test")

				So(update, ShouldBeTrue)
				So(len(steps), ShouldEqual, 2)
				So(nstep.Type, ShouldEqual, types.AlarmStepTypeSateCounter)
				So(nstep.StateCounter, ShouldNotBeNil)
				So(nstep.Author, ShouldEqual, "test")
				So(nstep.Message, ShouldEqual, "test")

				So(steps[0].Type, ShouldEqual, si.Type)
				So(steps[1].Type, ShouldEqual, nstep.Type)
			})
		})
	})
}

func TestAlarmStepCropOldData(t *testing.T) {
	Convey("Setup", t, func() {
		s, err := mongo.NewSession(mongo.Timeout)
		So(err, ShouldBeNil)

		c := s.DB(canopsis.DbName).C(alarm.AlarmCollectionName)
		_, err = c.RemoveAll(nil)
		So(err, ShouldBeNil)
		steps := make(types.AlarmSteps, 0)
		s1 := types.AlarmStep{
			Value:     1,
			Message:   "coucou",
			Author:    "coucou",
			Timestamp: types.CpsTime{Time: time.Now()},
		}
		stateCounter := types.NewPyMap()
		stateCounter[types.AlarmStepStateIncrease] = 1
		stateCounter[types.AlarmStepStateDecrease] = 2

		s2 := types.AlarmStep{
			Type:         types.AlarmStepTypeSateCounter,
			StateCounter: &stateCounter,
		}

		So(steps.Add(s1), ShouldBeNil)
		So(steps.Add(s2), ShouldBeNil)

		al := types.Alarm{
			ID: "cropalarm",
			Value: types.AlarmValue{
				Steps: steps,
			},
		}

		Convey("Inserting the alarm is not a problem", func() {
			So(c.Insert(al), ShouldBeNil)

			Convey("Getting the alarm back isn't a problem either", func() {
				var al types.Alarm
				So(c.FindId("cropalarm").One(&al), ShouldBeNil)

				So(len(al.Value.Steps), ShouldEqual, 2)
				So(al.Value.Steps[0].Value, ShouldEqual, 1)
				So(al.Value.Steps[0].StateCounter, ShouldBeNil)
				So(al.Value.Steps[1].Value, ShouldEqual, 0)
				So(al.Value.Steps[1].Type, ShouldEqual, types.AlarmStepTypeSateCounter)
				So(al.Value.Steps[1].StateCounter, ShouldNotBeNil)
				So(len(*al.Value.Steps[1].StateCounter), ShouldEqual, 2)

				sc := al.Value.Steps[1].StateCounter

				So(sc.Get(types.AlarmStepStateDecrease, 0), ShouldEqual, 2)
				So(sc.Get(types.AlarmStepStateIncrease, 0), ShouldEqual, 1)
			})
		})
	})
}
