package types

import (
	"errors"
	"fmt"
	"math/rand"
	"strconv"
	"strings"
	"time"

	"github.com/globalsign/mgo/bson"
)

// CpsNumber is here for compatibility with old python engines.
// It will force an int64 from a float64.
type CpsNumber int64

// MarshalJSON implements json.Encoder interface
func (t CpsNumber) MarshalJSON() ([]byte, error) {
	return []byte(fmt.Sprint(t)), nil
}

// UnmarshalJSON implements json.Decoder interface
func (t *CpsNumber) UnmarshalJSON(b []byte) error {
	f, err := strconv.ParseFloat(string(b), 64)
	*t = CpsNumber(f)
	return err
}

// GetBSON implements bson.Getter interface
func (t CpsNumber) GetBSON() (interface{}, error) {
	return int64(t), nil
}

// SetBSON implements bson.Setter interface
func (t *CpsNumber) SetBSON(raw bson.Raw) error {
	var num float64

	if err := raw.Unmarshal(&num); err != nil {
		return err
	}

	*t = CpsNumber(num)
	return nil
}

// Float64 ...
func (t CpsNumber) Float64() float64 {
	return float64(t)
}

// CpsTimestamp convert a number to a timestamp
func (t CpsNumber) CpsTimestamp() CpsTime {
	return CpsTime{time.Unix(int64(t), 0)}
}

// CpsTime allows conversion from time.Time to time.Time.Unix()
type CpsTime struct {
	time.Time
}

// MarshalJSON converts from CpsTime to timestamp as bytes
func (t CpsTime) MarshalJSON() ([]byte, error) {
	ts := time.Time(t.Time).Unix()
	stamp := fmt.Sprint(ts)

	return []byte(stamp), nil
}

// UnmarshalJSON converts from string to CpsTime
func (t *CpsTime) UnmarshalJSON(b []byte) error {
	ts, err := strconv.Atoi(string(b))
	if err != nil {
		tt, err := strconv.ParseFloat(string(b), 64)
		if err != nil {
			return err
		}
		ts = int(tt)
	}

	*t = CpsTime{time.Unix(int64(ts), 0)}
	return nil
}

// GetBSON implements the bson.Getter interface
func (t CpsTime) GetBSON() (interface{}, error) {
	ts := time.Time(t.Time).Unix()
	return ts, nil
}

// SetBSON implements the bson.Setter interface
func (t *CpsTime) SetBSON(raw bson.Raw) error {
	var ts int64

	if err := raw.Unmarshal(&ts); err != nil {
		return err
	}

	*t = CpsTime{time.Unix(int64(ts), 0)}
	return nil
}

// Format easilly format a CpsTime as a string
func (t CpsTime) Format() string {
	return t.Time.Format(time.RFC3339Nano)
}

// CpsDuration allow conversions from/to time.Duration to/from string
type CpsDuration time.Duration

// MarshalJSON converts a CpsDuration to string
func (t CpsDuration) MarshalJSON() ([]byte, error) {
	return []byte("\"" + time.Duration(t).String() + "\""), nil
}

// UnmarshalJSON converts a string to CpsDuration
func (t *CpsDuration) UnmarshalJSON(b []byte) error {
	sduration := string(b)

	if len(sduration) < 3 {
		return errors.New("Bad duration: string length below 3 chars")
	}

	if sduration[0] != '"' || sduration[len(sduration)-1] != '"' {
		return errors.New("Bad duration: not a string")
	}

	parsed, err := time.ParseDuration(sduration[1 : len(sduration)-1])
	if err != nil {
		return err
	}
	*t = CpsDuration(parsed)
	return nil
}

// GetBSON implements the bson.Getter interface
func (t CpsDuration) GetBSON() (interface{}, error) {
	return time.Duration(t).String(), nil
}

// SetBSON implements the bson.Setter interface
func (t *CpsDuration) SetBSON(raw bson.Raw) error {
	var d string

	if err := raw.Unmarshal(&d); err != nil {
		return err
	}

	parsed, err := time.ParseDuration(d)
	if err != nil {
		return err
	}
	*t = CpsDuration(parsed)
	return nil
}

// Duration return the CpsDuration casted to time.Duration
func (t CpsDuration) Duration() time.Duration {
	return time.Duration(t)
}

func listOfInterfaceToString(v []interface{}) (string, error) {
	values := make([]string, len(v))
	for i, vv := range v {
		sval, err := InterfaceToString(vv)
		if err != nil {
			return "", fmt.Errorf("list of: %v", err)
		}
		values[i] = sval
	}

	return strings.Join(values, ","), nil
}

// InterfaceToString tries to convert v to it's string value.
// Supported types:
// * float64
// * string
// * bool
// * int|int64|uint|uint64
// * []interface{} : join elements with ","
// * nil: will return empty string
//
// Any other type will return empty string and an error, like lists
// or maps...
func InterfaceToString(v interface{}) (string, error) {
	switch vt := v.(type) {
	case bool:
		return strconv.FormatBool(vt), nil
	case float64:
		return strconv.FormatFloat(vt, 'f', -1, 64), nil
	case string:
		return vt, nil
	case int:
		return strconv.FormatInt(int64(vt), 10), nil
	case int64:
		return strconv.FormatInt(vt, 10), nil
	case uint:
		return strconv.FormatUint(uint64(vt), 10), nil
	case uint64:
		return strconv.FormatUint(uint64(vt), 10), nil
	case []interface{}:
		return listOfInterfaceToString(vt)
	default:
		return "", fmt.Errorf("unsupported type: %T", v)
	}
}

const ascii_letters = "abcdefghijklmnopqrstuvwxyz"

// RandString generate a random string
func RandString(n int) string {
	b := make([]byte, n)
	for i := range b {
		b[i] = ascii_letters[rand.Intn(len(ascii_letters))]
	}
	return string(b)
}

// GenDisplayName generate a random uppercased display name
func GenDisplayName() string {
	name := RandString(2) + "-" + RandString(2) + "-" + RandString(2)
	return strings.ToUpper(name)
}
