package types

import (
	"encoding/json"
	"errors"
	"fmt"
	"reflect"
	"strings"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/errt"

	cps "git.canopsis.net/canopsis/go-revolution/lib/canopsis"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/heartbeat"
)

//Source types
const (
	SourceTypeResource  = "resource"
	SourceTypeComponent = "component"
)

//Event types
const (
	EventTypeAck           = "ack"
	EventTypeAckremove     = "ackremove"
	EventTypeAssocTicket   = "assocticket"
	EventTypeCalendar      = "calendar"
	EventTypeCancel        = "cancel"
	EventTypeCheck         = "check"
	EventTypeComment       = "comment"
	EventTypeConsolidation = "consolidation"
	EventTypeDeclareTicket = "declareticket"
	EventTypeDowntime      = "downtime"
	EventTypeEue           = "eue"
	EventTypeLog           = "log"
	EventTypeChangestate   = "changestate"
	EventTypeKeepstate     = "keepstate"
	EventTypePBehavior     = "pbehavior"
	EventTypePerf          = "perf"
	EventTypeSelector      = "selector"
	EventTypeSLA           = "sla"
	EventTypeSnooze        = "snooze"
	EventTypeTrap          = "trap"
	EventTypeUncancel      = "uncancel"
	EventTypeUser          = "user"
	EventTypeWatcher       = "watcher"
)

//PerfData represents a perf data array
type PerfData struct {
	Metric string  `bson:"metric" json:"metric"`
	Unit   string  `bson:"unit" json:"unit"`
	Value  float64 `bson:"value" json:"value"`
}

// Event represents a canopsis event.
type Event struct {
	ID            *string                `bson:"_id" json:"_id"`
	Connector     string                 `bson:"connector" json:"connector"`
	ConnectorName string                 `bson:"connector_name" json:"connector_name"`
	EventType     string                 `bson:"event_type" json:"event_type"`
	Component     string                 `bson:"component" json:"component"`
	Resource      string                 `bson:"resource" json:"resource"`
	PerfData      *string                `bson:"perf_data" json:"perf_data"`
	PerfDataArray []PerfData             `bson:"perf_data_array" json:"perf_data_array"`
	Status        *CpsNumber             `bson:"status" json:"status"`
	Timestamp     CpsTime                `bson:"timestamp" json:"timestamp"`
	StateType     *CpsNumber             `bson:"state_type" json:"state_type"`
	SourceType    string                 `bson:"source_type" json:"source_type"`
	LongOutput    string                 `bson:"long_output" json:"long_output"`
	State         CpsNumber              `bson:"state" json:"state"`
	Output        string                 `bson:"output" json:"output"`
	Alarm         *Alarm                 `bson:"current_alarm" json:"current_alarm"`
	Entity        *Entity                `bson:"current_entity" json:"current_entity"`
	Author        string                 `bson:"author" json:"author"`
	RK            string                 `bson:"routing_key" json:"routing_key"`
	MassAck       bool                   `json:"mass_ack"`
	Duration      *CpsNumber             `json:"duration"`
	Ticket        string                 `bson:"ticket" json:"ticket"`
	ExtraInfos    map[string]interface{} `json:"extra"`
	//CRecordType       string     `bson: json: `
	//CrecordWriteTime  *time.Time `bson: component json: component`
	//CrecordName       string     `bson: component json: component`
	//Enable            bool       `bson: component json: component`
	//Tags              []string   `bson: component json: component`
	//LastStateChange   *time.Time `bson:"" json:""`
}

// ContextInformation regroup context values necessary for creating a new entity
type ContextInformation struct {
	ID      string // Entity ID
	Name    string
	Type    string
	Impacts DedupList
	Depends DedupList
}

// NewEventFromJSON create an Event from a raw json
func NewEventFromJSON(body []byte) (Event, error) {
	var event Event
	err := json.Unmarshal(body, &event)
	if err != nil {
		return event, fmt.Errorf("NewEvent error: %v", err)
	}
	event.Format()

	return event, nil
}

// NewAlarmStep create a standard step from an event
func (e Event) NewAlarmStep(stepType string) AlarmStep {
	author := e.Author
	if e.Author == "" {
		author = cps.DefaultEventAuthor
	}
	if e.Timestamp.IsZero() {
		e.Timestamp = CpsTime{Time: time.Now()}
	}

	return AlarmStep{
		Author:    author,
		Message:   e.Output,
		Timestamp: e.Timestamp,
		Type:      stepType,
	}
}

// InjectExtraInfos takes the raw JSON event document and puts any unknown
// field into Event.ExtraInfos
func (e *Event) InjectExtraInfos(source []byte) error {
	unmatchedParams := make(map[string]interface{})
	if err := json.Unmarshal(source, &unmatchedParams); err == nil {
		for _, jsonKey := range e.GetRequiredKeys() {
			delete(unmatchedParams, jsonKey)
		}
	} else {
		return fmt.Errorf("Event.InjectExtraInfos json decode: %v", err)
	}
	e.ExtraInfos = unmatchedParams

	return nil
}

// Format an event
//  "timestamp" is fill with time.Now()
//  "event_type" is fill with EventTypeCheck
//  if "entity" is not null, "impacts" and "depends" are ensured to be initialized
func (e *Event) Format() {
	if e.Timestamp.IsZero() {
		e.Timestamp = CpsTime{time.Now()}
	}
	if e.EventType == "" {
		e.EventType = EventTypeCheck
	}
	if e.StateType == nil {
		statetype := CpsNumber(1)
		e.StateType = &statetype
	}

	if e.Entity != nil {
		e.Entity.EnsureInitialized()
	}
}

// GetEID generates an eid from an event
func (e *Event) GetEID() string {
	if e.Resource != "" {
		return e.Resource + "/" + e.Component
	}
	return e.Component
}

// IsValid checks if an Event is valid for Canopsis processing.
// the error returned, if any, is of type errt.UnknownError
func (e Event) IsValid() error {
	if e.Component == "" {
		return errt.NewUnknownError(errors.New("missing component"))
	}

	switch e.SourceType {
	case SourceTypeComponent:
		break
	case SourceTypeResource:
		if e.Resource == "" {
			return errt.NewUnknownError(errors.New("missing resource"))
		}
	default:
		return errt.NewUnknownError(fmt.Errorf("wrong source type: %v", e.SourceType))
	}

	switch e.EventType {
	case EventTypePerf:
		if e.PerfData == nil || e.PerfDataArray == nil {
			return errt.NewUnknownError(errors.New("perfdata without data"))
		}
	}

	return nil
}

// IsContextable tells you if the given event can lead to context enrichment.
func (e Event) IsContextable() bool {
	switch e.EventType {
	case EventTypeCheck:
		fallthrough
	case EventTypePerf:
		return true
	default:
		return false
	}
}

// GenericEvent contains an interface so you can do this:
// body := `<a json document>`
// var gevent GenericEvent
// json.Unmarshal(body, &gevent.Content)
// gevent.PartialID(<rules>)
type GenericEvent struct {
	Content interface{}
}

// JSONUnmarshal is a shortcut for this:
// var event GenericEvent
// json.Unmarshal(body, &event.Content)
func (e *GenericEvent) JSONUnmarshal(body []byte) error {
	return json.Unmarshal(body, &e.Content)
}

// PartialID builds the event ID, as a string, from the heartBeatItem configuration.
// Given this event:
//
// {
// 	"connector": "zabbix",
// 	"connector_name": "instance1",
// 	"component": "localhost",
// 	...
// }
//
// And the given heartBeatItem:
//
// li := NewHeartBeatItem(time.Minute*5)
// li.AddMapping("connector", "zabbix")
// li.AddMapping("connector_name", "instance1")
//
// PartialID will return "connector:zabbix.connector_name:instance1" as ID.
func (e *GenericEvent) PartialID(heartBeatItem heartbeat.Item) (string, error) {
	idvalues := make([]string, 0)

	if len(heartBeatItem.Mappings) == 0 {
		return "", errors.New("no mappings")
	}

	for _, mapping := range heartBeatItem.Mappings {
		for field := range mapping {
			switch event := e.Content.(type) {
			case map[string]interface{}:
				fieldValue, fexists := event[field]
				if !fexists {
					return "", fmt.Errorf("field %s does not exist", field)
				}
				sfieldValue, err := InterfaceToString(fieldValue)
				if err != nil {
					return "", fmt.Errorf("event partial id: %v", err)
				}
				idvalues = append(idvalues, field+":"+sfieldValue)
			default:
				return "", errors.New("cannot fetch data")
			}
			break
		}
	}

	return strings.Join(idvalues, "."), nil
}

// GetCompatRK returns the event routing key. For compatibility only with old engines.
func (e *Event) GetCompatRK() string {
	if e.SourceType == SourceTypeResource {
		return fmt.Sprintf(
			"%s.%s.%s.%s.%s.%s",
			e.Connector,
			e.ConnectorName,
			e.EventType,
			e.SourceType,
			e.Component,
			e.Resource,
		)
	}
	return fmt.Sprintf(
		"%s.%s.%s.%s.%s",
		e.Connector,
		e.ConnectorName,
		e.EventType,
		e.SourceType,
		e.Component,
	)
}

// GetRequiredKeys read all declared json tags in the struct
func (e *Event) GetRequiredKeys() []string {
	var values []string

	typeof := reflect.TypeOf(e).Elem()

	for i := 0; i < typeof.NumField(); i++ {
		field := typeof.Field(i)
		tag := field.Tag.Get("json")

		values = append(values, tag)
	}

	return values
}

// GenerateContextInformations generate connector, component and resource entity informations.
// First element is always connector
// Second element is always component
// Third element doesnt exists if event is not SourceTypeResource, otherwise it is resource
func (e *Event) GenerateContextInformations() []ContextInformation {
	connectorID := e.Connector + "/" + e.ConnectorName
	componentID := e.Component

	connector := ContextInformation{
		ID:      connectorID,
		Name:    e.ConnectorName,
		Type:    EntityTypeConnector,
		Impacts: NewDedupList(),
		Depends: NewDedupList(componentID),
	}

	component := ContextInformation{
		ID:      componentID,
		Name:    e.Component,
		Type:    EntityTypeComponent,
		Impacts: NewDedupList(connectorID),
		Depends: NewDedupList(),
	}

	if e.SourceType != SourceTypeResource {
		return []ContextInformation{connector, component}
	}

	resourceID := e.Resource + "/" + e.Component

	connector.Impacts.Add(resourceID)
	component.Depends.Add(resourceID)

	resource := ContextInformation{
		ID:      resourceID,
		Name:    e.Resource,
		Type:    EntityTypeResource,
		Impacts: NewDedupList(componentID),
		Depends: NewDedupList(connectorID),
	}

	return []ContextInformation{connector, component, resource}
}

// NewEntity create an entity from ContextInformation struct
func (ci *ContextInformation) NewEntity() Entity {
	return NewEntity(ci.ID, ci.Name, ci.Type, make(map[string]Info, 0), ci.Impacts, ci.Depends)
}
