package types_test

import (
	"encoding/json"
	"testing"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/heartbeat"
	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/types"
	. "github.com/smartystreets/goconvey/convey"
)

func TestPartialID(t *testing.T) {
	Convey("HeartBeatItem and Event", t, func() {
		li := heartbeat.NewItem(time.Second * 2)
		li.AddMapping("connector", "none")
		li.AddMapping("component", "none")

		bevent := []byte(`{"connector": "test", "component": "testc"}`)
		var ievent types.GenericEvent

		Convey("GenericEvent can JSONUnmarshal", func() {
			So(ievent.JSONUnmarshal(bevent), ShouldBeNil)

			Convey("I must have a valid PartialID without error", func() {
				id, err := ievent.PartialID(li)
				So(id, ShouldEqual, "connector:test.component:testc")
				So(err, ShouldBeNil)
			})
		})
	})
}

func TestEmptyPartialID(t *testing.T) {
	Convey("Empty PartialID", t, func() {
		li := heartbeat.NewItem(0)
		bevent := []byte(`{"connector":"test"}`)
		var ievent types.GenericEvent

		Convey("GenericEvent can JSONUnmarshal", func() {
			So(ievent.JSONUnmarshal(bevent), ShouldBeNil)

			Convey("GenericEvent can PartialID", func() {
				id, err := ievent.PartialID(li)
				So(id, ShouldEqual, "")
				So(err, ShouldNotBeNil)
			})
		})
	})
}

func TestBadJSONPartialID(t *testing.T) {
	Convey("Bad Partial ID", t, func() {
		var ievent types.GenericEvent
		sevent := `"bla"`
		li := heartbeat.NewItem(0)
		li.AddMapping("connector", "bla")

		Convey("GenericEvent can JSONUnmarshal", func() {
			err := ievent.JSONUnmarshal([]byte(sevent))
			So(err, ShouldBeNil)

			Convey("PartialID raises an error", func() {
				_, err = ievent.PartialID(li)

				So(err, ShouldNotBeNil)
				So(err.Error(), ShouldEqual, "cannot fetch data")
			})
		})

	})
}

func TestBadJSON2PartialID(t *testing.T) {
	Convey("Given a bad json document", t, func() {
		var ievent types.GenericEvent
		ev := []byte(`{"bla": ["bla"]}`)

		Convey("I must unmarshal it", func() {
			err := ievent.JSONUnmarshal(ev)
			So(err, ShouldBeNil)

			Convey("But i cannot get a partial id from a bad key", func() {
				li := heartbeat.NewItem(0)
				li.AddMapping("bla", "vert")
				id, err := ievent.PartialID(li)

				So(err, ShouldBeNil)
				So(id, ShouldEqual, "bla:bla")
				So(li.ID(), ShouldNotEqual, id)
				So(li.ID(), ShouldEqual, "bla:vert")
			})
		})
	})
}

func TestNoFieldPartialID(t *testing.T) {
	var ievent types.GenericEvent
	ev := []byte(`{"connector": "bla"}`)

	err := ievent.JSONUnmarshal(ev)
	if err != nil {
		t.Fatalf("json unmarshal: %v", err)
	}

	li := heartbeat.NewItem(0)
	li.AddMapping("ZBLEURRGANRSTUIEBL", "^_^")
	id, err := ievent.PartialID(li)
	if err == nil {
		t.Fatal("no error returned")
	}
	if id != "" {
		t.Fatalf("got non empty id: %s", id)
	}
}

func getEvent() types.Event {
	var event = types.Event{
		Connector:     "red",
		ConnectorName: "is",
		Component:     "dead",
		Resource:      "adieu_yourï",
		EventType:     types.EventTypeCheck,
		SourceType:    types.SourceTypeResource,
		LongOutput:    "",
		State:         types.AlarmStateMajor,
		Output:        "",
		PerfData:      nil,
		PerfDataArray: nil,
		ID:            nil,
		Status:        nil,
		StateType:     nil,
		Alarm:         nil,
		Entity:        nil,
	}
	event.Format()

	return event
}

// Generate a bad event
func GetBadEvent() types.Event {

	status := types.CpsNumber(-3)

	return types.Event{
		Connector:     "red",
		ConnectorName: "is",
		Component:     "nil !",
		Resource:      "",
		EventType:     "wrong_event_type",
		PerfData:      nil,
		PerfDataArray: nil,
		ID:            nil,
		Status:        &status,
		StateType:     nil,
		SourceType:    "wrong_source_type",
		LongOutput:    "",
		State:         -1,
		Output:        "",
		Alarm:         nil,
		Entity:        nil,
	}
}

func GetBadEvent2() types.Event {
	evt := GetBadEvent()
	evt.EventType = ""
	return evt
}

func TestEventTypeIsReplaced(t *testing.T) {
	Convey("Given an event with EventType empty", t, func() {
		evt := GetBadEvent2()

		Convey("EventType goes to EventTypeCheck when Format() is called", func() {
			evt.Format()
			So(evt.EventType, ShouldEqual, types.EventTypeCheck)
		})
	})
}

func TestEventIsGood(t *testing.T) {

	Convey("Given an event", t, func() {
		e := getEvent()

		Convey("The event is valid", func() {
			So(e.IsValid(), ShouldBeNil)
		})

		Convey("The event is correctly formatted", func() {
			e.Format()
			So(e.Timestamp, ShouldNotBeNil)
			So(e.EventType, ShouldNotBeBlank)
		})
	})
}

func TestEventIsBad(t *testing.T) {

	Convey("Given a bad event", t, func() {
		e := GetBadEvent()

		Convey("The event is not valid", func() {
			So(e.IsValid(), ShouldNotBeNil)
		})

		Convey("The event is correctly formatted anyway", func() {
			e.Format()
			So(e.Timestamp, ShouldNotBeNil)
			So(e.EventType, ShouldNotBeNil)
			So(e.EventType, ShouldNotBeBlank)
			So(e.Output, ShouldNotBeNil)
		})
	})

	Convey("Given an event without resource", t, func() {
		e := getEvent()
		e.Resource = ""

		Convey("The event is not valid", func() {
			So(e.IsValid(), ShouldNotBeNil)
		})
	})

	Convey("Given a malformed perf event", t, func() {
		e := getEvent()
		e.EventType = types.EventTypePerf

		Convey("The event is not valid", func() {
			So(e.IsValid(), ShouldNotBeNil)
		})
	})
}

func TestListTypeFields(t *testing.T) {
	Convey("Given an event struct", t, func() {
		event := getEvent()
		Convey("Then basic fields are listed by listTypeFields()", func() {
			tags := event.GetRequiredKeys()
			So(tags, ShouldContain, "connector")
			So(tags, ShouldContain, "connector_name")
			So(tags, ShouldContain, "component")
		})
	})
}

func TestGenerateContextIds(t *testing.T) {

	Convey("Given an event", t, func() {
		e := getEvent()

		Convey("The generated context informations are corrects", func() {
			infos := e.GenerateContextInformations()
			So(infos[0].ID, ShouldEqual, "red/is")
			So(infos[1].ID, ShouldEqual, "dead")
			So(infos[2].ID, ShouldEqual, "adieu_yourï/dead")
			So(infos[2].Type, ShouldEqual, types.SourceTypeResource)
		})

		Convey("The generated context informations are corrects, event without resource", func() {
			e.SourceType = types.SourceTypeComponent
			e.Resource = ""

			infos := e.GenerateContextInformations()
			So(infos[0].ID, ShouldEqual, "red/is")
			So(infos[0].Impacts.List(), ShouldResemble, []string{})
			So(infos[0].Depends.List(), ShouldResemble, []string{"dead"})
			So(infos[1].ID, ShouldEqual, "dead")
			So(infos[1].Type, ShouldEqual, types.SourceTypeComponent)
			So(infos[1].Impacts.List(), ShouldResemble, []string{"red/is"})
			So(infos[1].Depends.List(), ShouldResemble, []string{})
		})
	})
}

var ctxInfos []types.ContextInformation

func BenchmarkGenerateContextInformations(b *testing.B) {
	e := getEvent()
	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		ctxInfos = e.GenerateContextInformations()
		if len(ctxInfos) == 0 {
			b.Fatal("no infos")
		}
	}
}

func TestEmptyEntity(t *testing.T) {
	Convey("Setup incomplete event", t, func() {
		evt := `
	{
		"event_type":"check",
		"component":"bla",
		"resource":"prout",
		"state":3,
		"current_entity":{},
		"connector":"bla",
		"connector_name":"bla"
	}
	`

		var e types.Event

		So(json.Unmarshal([]byte(evt), &e), ShouldBeNil)
		So(e.Entity, ShouldNotBeNil)

		So(func() {
			e.Entity.Depends.Add("bla")
		}, ShouldPanic)

		So(func() {
			e.Entity.Impacts.Add("bla")
		}, ShouldPanic)

		Convey("Format event", func() {
			e.Format()
			So(func() {
				e.Entity.Depends.Add("bla")
			}, ShouldNotPanic)

			So(func() {
				e.Entity.Impacts.Add("bla")
			}, ShouldNotPanic)
		})

	})
}

func TestInjectExtraInfos(t *testing.T) {
	Convey("Setup", t, func() {
		evt := `{
		"event_type":"check",
		"component":"bla",
		"resource":"blurk",
		"state":3,
		"connector":"bla",
		"connector_name":"bla",
		"personnemeconnait":"ulyss31",
		"personnemeconnait2":"ulyss62"
	}`
		var e types.Event
		So(json.Unmarshal([]byte(evt), &e), ShouldBeNil)

		Convey("Extra informations are into ExtraInfos", func() {
			So(e.InjectExtraInfos([]byte(evt)), ShouldBeNil)
			So(e.ExtraInfos, ShouldContainKey, "personnemeconnait")
			So(e.ExtraInfos, ShouldContainKey, "personnemeconnait2")
			So(e.ExtraInfos["personnemeconnait"], ShouldEqual, "ulyss31")
			So(e.ExtraInfos["personnemeconnait2"], ShouldEqual, "ulyss62")
		})
	})
}
