package types

import (
	"errors"
	"fmt"
	"log"
	"time"

	"git.canopsis.net/canopsis/go-revolution/lib/canopsis/config"

	"github.com/google/uuid"
)

//Alarm states
const (
	AlarmStateOK = iota
	AlarmStateMinor
	AlarmStateMajor
	AlarmStateCritical
)

// Alarm state types
const (
	AlarmStateTypeChangeState = "changestate"
)

//Alarm statuses
const (
	AlarmStatusOff = iota
	AlarmStatusOngoing
	AlarmStatusFlapping
	AlarmStatusStealthy
	AlarmStatusCancelled
)

//Alarm steps
const (
	AlarmStepStateIncrease  = "stateinc"
	AlarmStepStateDecrease  = "statedec"
	AlarmStepStatusIncrease = "statusinc"
	AlarmStepStatusDecrease = "statusdec"
	AlarmStepAck            = "ack"
	AlarmStepAckRemove      = "ackremove"
	AlarmStepCancel         = "cancel"
	AlarmStepUncancel       = "uncancel"
	AlarmStepComment        = "comment"
	AlarmStepDeclareTicket  = "declareticket"
	AlarmStepAssocTicket    = "assocticket"
	AlarmStepSnooze         = "snooze"
	AlarmStepPbehavior      = "pbehavior"
)

// Alarm represents an alarm document.
type Alarm struct {
	ID       string     `bson:"_id" json:"_id"`
	Time     CpsTime    `bson:"t" json:"t"`
	EntityID string     `bson:"d" json:"d"`
	Value    AlarmValue `bson:"v" json:"v"`
	isDirty  bool
}

// NewAlarmID creates a uuid
func NewAlarmID() string {
	return uuid.New().String()
}

// NewAlarm creates en new Alarm from an Event
func NewAlarm(event Event) (Alarm, error) {
	now := CpsTime{time.Now()}
	var alarmsteps AlarmSteps

	if event.Timestamp.IsZero() {
		return Alarm{}, errors.New("field Timestamp is not set")
	}

	alarm := Alarm{
		EntityID: event.GetEID(),
		ID:       NewAlarmID(),
		Time:     now,
		isDirty:  true,
		Value: AlarmValue{
			Connector:      event.Connector,
			ConnectorName:  event.ConnectorName,
			Component:      event.Component,
			Resource:       event.Resource,
			Extra:          map[string]string{},
			Tags:           make([]string, 0),
			Steps:          alarmsteps,
			CreationDate:   now,
			LastUpdateDate: now,
			LastEventDate:  now,
			DisplayName:    GenDisplayName(),
		},
	}
	alarm.Update(event)

	return alarm, nil
}

// AlarmID build an alarmid from given parameters. Used by Alarm.AlarmID()
func AlarmID(connector, connectorName, entityID string) string {
	return fmt.Sprintf(
		"%s/%s/%s",
		connector,
		connectorName,
		entityID,
	)
}

// AlarmID returns current alarm's alarmid.
func (a *Alarm) AlarmID() string {
	return AlarmID(
		a.Value.Connector,
		a.Value.ConnectorName,
		a.EntityID,
	)
}

// CacheID implements cache.Cache interface
func (a Alarm) CacheID() string {
	return a.AlarmID()
}

// CropSteps calls Crop() on Alarm.Value.Steps with alarm parameters.
// returns true if the alarm was modified.
func (a *Alarm) CropSteps() bool {
	updated := false
	if a.Value.State != nil {
		cropStep, cropUpdate := a.Value.Steps.Crop(
			a.Value.State.Value,
			AlarmStepCropMinStates,
			a.Value.State.Author,
			a.Value.State.Message,
		)
		if cropUpdate {
			a.Value.State = &cropStep
		}

		updated = updated || cropUpdate
	}
	return updated
}

// GetCurrentState returns the Current State of the Alarm
func (a *Alarm) GetCurrentState() CpsNumber {
	if a.Value.State == nil {
		return AlarmStateOK
	}
	lastAlarmDate := a.Value.State.Value

	if lastAlarmDate < AlarmStateOK {
		return AlarmStateOK
	}

	return lastAlarmDate
}

// CurrentStatus returns the Current status of the alarm
func (a *Alarm) CurrentStatus() CpsNumber {
	if a.Value.Status == nil {
		return AlarmStatusOff
	}
	currentAlarmStatus := a.Value.Status.Value
	if currentAlarmStatus < AlarmStatusOff {
		return AlarmStatusOff
	}

	return currentAlarmStatus
}

// Update an alarm from an Event
func (a *Alarm) Update(e Event) {
	a.updateState(e)
	a.updateStatus(e)
}

// updateState If Event is a check event and if the associated alarm is not
// locked, increases or decreases the State of the alarm.
func (a *Alarm) updateState(e Event) {
	if e.EventType != EventTypeCheck {
		return
	}

	currentAlarmState := a.GetCurrentState()
	if e.State == currentAlarmState {
		// State hasn't changed. Stop here
		return
	}
	if a.IsStateLocked() {
		// Event is an OK, so the alarm should be resolved anyway
		if e.State == AlarmStateOK {
			a.Value.State.Type = ""
			a.isDirty = true
		} else {
			// Stop here: state should be preserved.
			return
		}
	}

	// Create new Step to keep track of the alarm history
	newStep := AlarmStep{
		Type:      AlarmStepStateIncrease,
		Timestamp: CpsTime(e.Timestamp),
		Author:    e.Connector + "." + e.ConnectorName,
		Message:   e.Output,
		Value:     e.State,
	}
	if e.State < currentAlarmState {
		newStep.Type = AlarmStepStateDecrease
	}

	a.Value.State = &newStep
	a.Value.Steps.Add(newStep)
	a.isDirty = true
}

// Update the status of an Alarm [sic]
func (a *Alarm) updateStatus(e Event) {
	currentAlarmStatus := a.CurrentStatus()
	newStatus := a.ComputeStatus()
	if newStatus == currentAlarmStatus {
		return
	}

	// Create new Step to keep track of the alarm history.
	newStep := AlarmStep{
		Type:      AlarmStepStatusIncrease,
		Timestamp: CpsTime(e.Timestamp),
		Author:    e.Connector + "." + e.ConnectorName,
		Message:   e.Output,
		Value:     newStatus,
	}

	// If the status is Decreasing
	if newStatus < currentAlarmStatus {
		newStep.Type = AlarmStepStatusDecrease
	}
	a.Value.Status = &newStep
	a.Value.Steps.Add(newStep)
}

// ComputeStatus of an Alarm from an Event
func (a *Alarm) ComputeStatus() CpsNumber {
	if a.IsCanceled() {
		return AlarmStatusCancelled
	}

	if a.IsFlapping() {
		return AlarmStatusFlapping
	}

	if a.IsStealthy() {
		return AlarmStatusStealthy
	}

	if a.Value.State.Value != AlarmStateOK {
		return AlarmStatusOngoing
	}

	return AlarmStatusOff
}

// Ack an alarm
func (a *Alarm) Ack(event Event) AlarmStep {
	newStep := event.NewAlarmStep(AlarmStepAck)
	a.Value.ACK = &newStep

	return newStep
}

// Unack removes an ack on an alarm
func (a *Alarm) Unack(event Event) AlarmStep {
	newStep := event.NewAlarmStep(AlarmStepAckRemove)
	a.Value.ACK = nil

	return newStep
}

// Cancel cancel an alarm
func (a *Alarm) Cancel(event Event) AlarmStep {
	newStep := event.NewAlarmStep(AlarmStepCancel)
	a.Value.Canceled = &newStep

	return newStep
}

// Uncancel uncancel an alarm
func (a *Alarm) Uncancel(event Event) AlarmStep {
	newStep := event.NewAlarmStep(AlarmStepUncancel)
	a.Value.Canceled = nil

	return newStep
}

// ChangeState force the state of an alarm
func (a *Alarm) ChangeState(event Event) AlarmStep {
	newStep := event.NewAlarmStep(AlarmStateTypeChangeState)
	newStep.Value = event.State
	a.Value.State = &newStep

	return newStep
}

// Comment comment an alarm
func (a *Alarm) Comment(event Event) AlarmStep {
	newStep := event.NewAlarmStep(AlarmStepComment)
	// It just add a step with a comment

	return newStep
}

// Ticket add a ticket on an alarm
func (a *Alarm) Ticket(stepType string, event Event) AlarmStep {
	newStep := event.NewAlarmStep(stepType)
	newStep.Message = event.Ticket

	ticketStep := newStep.NewTicket(event.Ticket)
	a.Value.Ticket = &ticketStep

	return newStep
}

// AssocTicket associate a ticket number to an alarm
func (a *Alarm) AssocTicket(event Event) AlarmStep {
	return a.Ticket(AlarmStepAssocTicket, event)
}

// DeclareTicket ask for a creation
func (a *Alarm) DeclareTicket(event Event) AlarmStep {
	//TODO: declare should  not read ticket parameter, but schedule a ticket
	// creation and get back a ticket number.
	// ! BUT ! On "ack and report" action from frontend, it send a declareticket
	// event with the corresponding ticket number...
	return a.Ticket(AlarmStepDeclareTicket, event)
}

// Resolve mark as resolved an Alarm with a timestamp [sic]
func (a *Alarm) Resolve(timestamp *CpsTime) {
	a.Value.Resolved = timestamp
}

// ResolveCancel forces alarm resolution on cancel
func (a *Alarm) ResolveCancel(timestamp *CpsTime) {
	a.Resolve(timestamp)
	a.Value.Status = a.Value.Canceled
	a.Value.Status.Value = AlarmStatusCancelled
}

// Snooze apply a snooze step to an Alarm
func (a *Alarm) Snooze(event Event) AlarmStep {
	newStep := event.NewAlarmStep(AlarmStepSnooze)
	newStep.Value = CpsNumber(event.Timestamp.Time.Unix()) + *event.Duration
	a.Value.Snooze = &newStep

	return newStep
}

// UnSnooze cancel a snooze
func (a *Alarm) UnSnooze() {
	a.Value.LastUpdateDate = CpsTime{time.Now()}
	a.Value.Snooze = nil
}

// Closable checks the last step for it's state to be OK for at least
// baggotTime. Reference time is time.Now() when this function is called.
func (a Alarm) Closable(baggotTime time.Duration) bool {
	// prevent some silly crash
	if a.Value.State == nil {
		return false
	}

	ls, err := a.Value.Steps.Last()
	if err == nil && a.Value.State.Value == AlarmStateOK {
		return true
	}

	if err != nil && a.Value.State.Value != AlarmStateOK {
		log.Printf("warning: alarm %s has empty steps but is not OK", a.ID)
		return false
	}

	timeok := ls.Timestamp.Time.Before(time.Now().Add(-baggotTime))
	if a.Value.State.Value == AlarmStateOK && timeok {
		return true
	}

	return false
}

// IsAck check if an Alarm is acked
func (a Alarm) IsAck() bool {
	return a.Value.ACK != nil // && !a.IsResolved()
}

// IsCanceled check if an Alarm is canceled
func (a Alarm) IsCanceled() bool {
	return a.Value.Canceled != nil && !a.IsResolved()
}

// IsDirty indicate if an Alarm has been modified
func (a Alarm) IsDirty() bool {
	return a.isDirty
}

// IsFlapping check if an Alarm is currently flapping
func (a Alarm) IsFlapping() bool {
	conf := config.GetConf()
	lastStepType := ""
	freq := 0

	for i := len(a.Value.Steps) - 1; i >= 0; i-- {
		s := a.Value.Steps[i]
		duration := int(time.Since(s.Timestamp.Time).Seconds())
		if duration >= conf.Alarm.FlappingInterval {
			break
		}

		if s.Type != lastStepType {
			switch s.Type {
			case AlarmStepStateIncrease:
				fallthrough
			case AlarmStepStateDecrease:
				lastStepType = s.Type
				freq++
			}
		}

		if freq > conf.Alarm.FlappingFreqLimit {
			return true
		}
	}

	return false
}

// IsResolved tell if an alarm has been resolved
func (a *Alarm) IsResolved() bool {
	return a.Value.Resolved != nil
}

// IsSnoozed check if an Alarm is snoozed
func (a Alarm) IsSnoozed() bool {
	if a.Value.Snooze == nil {
		return false
	}

	snoozeEnd := a.Value.Snooze.Value.CpsTimestamp()
	return snoozeEnd.After(time.Now())
}

// IsStateLocked checks that the Alarm is not Locked (by manual intervention for example)
func (a *Alarm) IsStateLocked() bool {
	return a.Value.State != nil && a.Value.State.Type == AlarmStateTypeChangeState
}

// IsStealthy checks if an Alarm is currently stealthy
func (a Alarm) IsStealthy() bool {
	// FIXME: crash on watch event creating a new alarm, cause State is not initialized
	if a.Value.State.Value != AlarmStateOK {
		return false
	}
	conf := config.GetConf()

	for i := len(a.Value.Steps) - 1; i >= 0; i-- {
		s := a.Value.Steps[i]
		timeSinceThisStep := time.Since(s.Timestamp.Time)
		if int(timeSinceThisStep.Seconds()) >= conf.Alarm.StealthyInterval {
			break
		}

		if s.Value != AlarmStateOK {
			switch s.Type {
			case AlarmStepStatusIncrease:
				fallthrough
			case AlarmStepStateDecrease:
				return true
			default:
				break
			}
		}
	}

	return false
}
