package engine

import (
	"os"
	"os/signal"
	"syscall"
)

// FullEngine is a Worker + Periodical engine
type FullEngine struct {
	Main
	Worker
	Periodical
}

// DefaultEngine is the default Engine implementation
type DefaultEngine struct {
	stopAsked   bool
	stopSignal  chan os.Signal
	stopped     chan bool
	cont        bool
	returnValue int
}

// NewDefaultEngine creates an engine that handles SIGINT and SIGTERM
// as stop signals.
// It only implements engine.Main.
func NewDefaultEngine() *DefaultEngine {
	e := DefaultEngine{
		stopAsked:   false,
		stopSignal:  make(chan os.Signal, 1),
		stopped:     make(chan bool, 2),
		cont:        true,
		returnValue: ExitOK,
	}

	return &e
}

// Initialize setup SIGINT, SIGTERM and SIGKILL catch and must be called
// by other Initialize() functions.
// Any error != nil returned will prevent the engine to be started.
func (e *DefaultEngine) Initialize() error {
	signal.Notify(e.stopSignal, syscall.SIGINT, syscall.SIGTERM)
	return nil
}

// Continue ...
func (e *DefaultEngine) Continue() bool {
	return e.cont
}

// ProcessStopped must not be called more than once per process that has
// effectively stopped.
func (e *DefaultEngine) ProcessStopped() {
	e.stopped <- true
}

// StopSignal returns the channel where signals are captured to handle
// graceful stop.
func (e *DefaultEngine) StopSignal() <-chan os.Signal {
	return e.stopSignal
}

// AskStop unlock engine.Start() wait on StopSignal()
// and set the engine to be stopped.
func (e *DefaultEngine) AskStop(state int) {
	if !e.stopAsked {
		e.returnValue = state
		e.stopSignal <- syscall.SIGINT
		e.stopAsked = true
		e.cont = false
	}
}

// WaitStopped awaits all processes to finish.
func (e *DefaultEngine) WaitStopped() {
	for i := 0; i < cap(e.stopped); i++ {
		<-e.stopped
	}
}
