package engine

import (
	"fmt"
	"os"
	"time"
)

// Returned values on exit
const (
	ExitOK = iota
	ExitPanic
	ExitEngine
)

// Main ...
type Main interface {
	Initialize() error
	Continue() bool
	// AskStop must change Continue() return value to false
	// and StopSignal() to deliver a value through the channel.
	AskStop(int)
	StopSignal() <-chan os.Signal
	// ProcessStopped must be called by each process when they are
	// actually stopped, or have recovered from a panic.
	ProcessStopped()
	WaitStopped()
}

func startWorker(w Worker) error {
	channel, err := w.WorkerConsumer()
	if err != nil {
		w.WorkerStopped()
		return fmt.Errorf("start worker: %v", err)
	}

	go func() {
		timeout := make(chan bool, 1)
		go func() {
			for w.WorkerContinue() {
				time.Sleep(time.Second)
				timeout <- true
			}
		}()

		defer w.WorkerRecover()
		for w.WorkerContinue() {
			select {
			case msg := <-channel:
				if len(msg.Body) > 0 {
					w.WorkerProcess(msg)
				}
			case <-timeout:
			}
		}

		w.WorkerStopped()
	}()

	return nil
}

func startPeriodical(p Periodical) {
	// watch engine.Continue every 2 second, during a maximum amount of time
	// given by engine.PeriodicalWaitTime(), and return engine.Continue().
	waitAndCheck := func() bool {
		sleepCountDown := p.PeriodicalWait()
		for sleepCountDown > 0 && p.PeriodicalContinue() {
			wait := time.Second * 1
			time.Sleep(wait)
			sleepCountDown = sleepCountDown - wait
		}
		return p.PeriodicalContinue()
	}

	go func() {
		defer p.PeriodicalRecover()
		for waitAndCheck() {
			//for engine.Continue() {
			p.PeriodicalProcess()
		}
		p.PeriodicalStopped()
	}()
}

// Start all subprocesses of an engine.
func Start(e Main) error {
	w, worker := e.(Worker)
	p, periodical := e.(Periodical)

	if !worker && !periodical {
		return fmt.Errorf("engine does not implement any of Worker or Periodical interfaces")
	}

	if err := e.Initialize(); err != nil {
		return fmt.Errorf("engine init error: %v", err)
	}

	if worker {
		if err := startWorker(w); err != nil {
			return fmt.Errorf("worker start: %v", err)
		}
	} else {
		e.ProcessStopped()
	}

	if periodical {
		startPeriodical(p)
	} else {
		e.ProcessStopped()
	}

	<-e.StopSignal()
	e.AskStop(ExitOK)
	e.WaitStopped()

	return nil
}
