package redis

import (
	"errors"
	"fmt"
	"net/url"
	"os"
	"strconv"

	"git.canopsis.net/canopsis/go-revolution/lib/utils"
	"github.com/go-redis/redis"
)

// Cache<Type> gives you constants to use for different caches.
const (
	CacheAlarm = iota
	CacheEntity
)

// Env vars for redis session
const (
	EnvURL = "CPS_REDIS_URL"
)

// NewOptions handles redis.Options creation based on
// the surl, which must be on the following shape:
//
// redis://[nouser:password@]host:port/int
// int must be un number indicating the database
//
// If you have a password for the database, no user is required.
// But it is required to not leave the user empty to avoid url
// parsing error.
func NewOptions(surl string, db int) (redis.Options, error) {
	redisURL, err := url.ParseRequestURI(surl)
	if err != nil {
		return redis.Options{}, err
	}

	redisPassword := ""
	redisPasswordSet := false
	if redisURL.User != nil {
		redisPassword, redisPasswordSet = redisURL.User.Password()
		if !redisPasswordSet {
			redisPassword = ""
		}
	}

	redisDB := db
	if db < 0 {
		if len(redisURL.Path) < 2 {
			return redis.Options{}, errors.New("no database specified in url")
		}
		rdb, err := strconv.ParseInt(redisURL.Path[1:], 10, 32)
		if err != nil {
			return redis.Options{}, err
		}

		redisDB = int(rdb)
	}

	redisOptions := redis.Options{
		Addr:     redisURL.Host,
		Password: redisPassword,
		DB:       redisDB,
	}

	return redisOptions, nil
}

// NewSession creates a new connection to a Redis database.
// Configuration is base on EnvCpsRedisUrl.
func NewSession(db int) (*redis.Client, error) {
	url := os.Getenv(EnvURL)
	if url == "" {
		return nil, fmt.Errorf("environment variable %s empty", EnvURL)
	}
	redisOptions, err := NewOptions(url, db)
	if err != nil {
		return nil, err
	}
	redisClient := redis.NewClient(&redisOptions)
	err = redisClient.Ping().Err()
	if err != nil {
		return nil, err
	}
	return redisClient, nil
}

// NewSessionFail calls NewSession and utils.ExitOnError
func NewSessionFail(db int) *redis.Client {
	s, err := NewSession(db)
	utils.ExitOnError(err, "redis session")
	return s
}
